# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

from unittest import TestCase

import numpy as np

import astronomia.globals
from astronomia.calendar import cal_to_jd, hms_to_fday
from astronomia.constants import days_per_second
from astronomia.equinox import equinox, equinox_approx

_months = {"spring": 3, "summer": 6, "autumn": 9, "winter": 12}


class TestEquinox(TestCase):
    def test_equinox(self):
        jd = equinox_approx(1962, "summer")
        self.assertAlmostEqual(jd, 2437837.39245, places=5)

        jd = equinox(2437837.38589, "summer", days_per_second)
        np.testing.assert_array_almost_equal(
            jd, cal_to_jd(1962, 6, 21) + hms_to_fday(21, 24, 42), decimal=5
        )

        tbl = [
            (
                1996,
                (
                    ("spring", 20, hms_to_fday(8, 4, 7)),
                    ("summer", 21, hms_to_fday(2, 24, 46)),
                    ("autumn", 22, hms_to_fday(18, 1, 8)),
                    ("winter", 21, hms_to_fday(14, 6, 56)),
                ),
            ),
            (
                1997,
                (
                    ("spring", 20, hms_to_fday(13, 55, 42)),
                    ("summer", 21, hms_to_fday(8, 20, 59)),
                    ("autumn", 22, hms_to_fday(23, 56, 49)),
                    ("winter", 21, hms_to_fday(20, 8, 5)),
                ),
            ),
            (
                1998,
                (
                    ("spring", 20, hms_to_fday(19, 55, 35)),
                    ("summer", 21, hms_to_fday(14, 3, 38)),
                    ("autumn", 23, hms_to_fday(5, 38, 15)),
                    ("winter", 22, hms_to_fday(1, 57, 31)),
                ),
            ),
            (
                1999,
                (
                    ("spring", 21, hms_to_fday(1, 46, 53)),
                    ("summer", 21, hms_to_fday(19, 50, 11)),
                    ("autumn", 23, hms_to_fday(11, 32, 34)),
                    ("winter", 22, hms_to_fday(7, 44, 52)),
                ),
            ),
            (
                2000,
                (
                    ("spring", 20, hms_to_fday(7, 36, 19)),
                    ("summer", 21, hms_to_fday(1, 48, 46)),
                    ("autumn", 22, hms_to_fday(17, 28, 40)),
                    ("winter", 21, hms_to_fday(13, 38, 30)),
                ),
            ),
            (
                2001,
                (
                    ("spring", 20, hms_to_fday(13, 31, 47)),
                    ("summer", 21, hms_to_fday(7, 38, 48)),
                    ("autumn", 22, hms_to_fday(23, 5, 32)),
                    ("winter", 21, hms_to_fday(19, 22, 34)),
                ),
            ),
            (
                2002,
                (
                    ("spring", 20, hms_to_fday(19, 17, 13)),
                    ("summer", 21, hms_to_fday(13, 25, 29)),
                    ("autumn", 23, hms_to_fday(4, 56, 28)),
                    ("winter", 22, hms_to_fday(1, 15, 26)),
                ),
            ),
            (
                2003,
                (
                    ("spring", 21, hms_to_fday(1, 0, 50)),
                    ("summer", 21, hms_to_fday(19, 11, 32)),
                    ("autumn", 23, hms_to_fday(10, 47, 53)),
                    ("winter", 22, hms_to_fday(7, 4, 53)),
                ),
            ),
            (
                2004,
                (
                    ("spring", 20, hms_to_fday(6, 49, 42)),
                    ("summer", 21, hms_to_fday(0, 57, 57)),
                    ("autumn", 22, hms_to_fday(16, 30, 54)),
                    ("winter", 21, hms_to_fday(12, 42, 40)),
                ),
            ),
            (
                2005,
                (
                    ("spring", 20, hms_to_fday(12, 34, 29)),
                    ("summer", 21, hms_to_fday(6, 47, 12)),
                    ("autumn", 22, hms_to_fday(22, 24, 14)),
                    ("winter", 21, hms_to_fday(18, 36, 1)),
                ),
            ),
        ]

        for yr, terms in tbl:
            for season, day, fday in terms:
                approx = equinox_approx(yr, season)
                jd = equinox(approx, season, days_per_second)
                np.testing.assert_array_almost_equal(
                    jd, cal_to_jd(yr, _months[season], day + fday), decimal=4
                )

    def test_equinox_range(self):
        """
        Check the accuracy of the equinox approximation routines over
        4000 years.

        Meeus provides formulae for appromimate solstices and equinoxes for
        the years -1000 to 3000. How accurate are they over the whole range
        of years?

        The test below compares the approximate values with the exact
        values as determined by the VSOP87d theory.

        In other tests, the maximum differences is 0.0015 days, or about 2.16
        minutes. The maximum occurred for the summer solstice in -408.

        """
        for yr in range(-1000, 3001, 100):
            for season in astronomia.globals.season_names:
                approx_jd = equinox_approx(yr, season)
                #
                # We use the 21st of the month as our guess, just in case the
                # approx_jd is wildly off.
                #
                jd = equinox(
                    cal_to_jd(yr, _months[season], 21), season, days_per_second
                )
                np.testing.assert_array_almost_equal(jd, [approx_jd], decimal=1)



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Functions to calculate nutation and obliquity values.

The IAU "1980 Theory of Nutation" is used, but terms with coefficients smaller
than 0.0003" have been dropped.

Reference: Jean Meeus, _Astronomical Algorithms_, second edition, 1998,
Willmann-Bell, Inc.

The first edition of the Meeus book had some errors in the table. These may be
corrected in the second edition. I recall correcting my values from those
published in _Explanatory Supplement to the Astronomical Almanac_, revised
edition edited by P. Kenneth Seidelman, 1992

"""

import numpy as np

from .calendar import jd_to_jcent
from .commonterms import kD, kF, kM, kM1, ko
from .util import d_to_r, dms_to_d, modpi2, polynomial

# [Meeus-1998: table 22.A]
#
#    D, M, M1, F, omega, psiK, psiT, epsK, epsT

_tbl = (
    (0, 0, 0, 0, 1, -171996, -1742, 92025, 89),
    (-2, 0, 0, 2, 2, -13187, -16, 5736, -31),
    (0, 0, 0, 2, 2, -2274, -2, 977, -5),
    (0, 0, 0, 0, 2, 2062, 2, -895, 5),
    (0, 1, 0, 0, 0, 1426, -34, 54, -1),
    (0, 0, 1, 0, 0, 712, 1, -7, 0),
    (-2, 1, 0, 2, 2, -517, 12, 224, -6),
    (0, 0, 0, 2, 1, -386, -4, 200, 0),
    (0, 0, 1, 2, 2, -301, 0, 129, -1),
    (-2, -1, 0, 2, 2, 217, -5, -95, 3),
    (-2, 0, 1, 0, 0, -158, 0, 0, 0),
    (-2, 0, 0, 2, 1, 129, 1, -70, 0),
    (0, 0, -1, 2, 2, 123, 0, -53, 0),
    (2, 0, 0, 0, 0, 63, 0, 0, 0),
    (0, 0, 1, 0, 1, 63, 1, -33, 0),
    (2, 0, -1, 2, 2, -59, 0, 26, 0),
    (0, 0, -1, 0, 1, -58, -1, 32, 0),
    (0, 0, 1, 2, 1, -51, 0, 27, 0),
    (-2, 0, 2, 0, 0, 48, 0, 0, 0),
    (0, 0, -2, 2, 1, 46, 0, -24, 0),
    (2, 0, 0, 2, 2, -38, 0, 16, 0),
    (0, 0, 2, 2, 2, -31, 0, 13, 0),
    (0, 0, 2, 0, 0, 29, 0, 0, 0),
    (-2, 0, 1, 2, 2, 29, 0, -12, 0),
    (0, 0, 0, 2, 0, 26, 0, 0, 0),
    (-2, 0, 0, 2, 0, -22, 0, 0, 0),
    (0, 0, -1, 2, 1, 21, 0, -10, 0),
    (0, 2, 0, 0, 0, 17, -1, 0, 0),
    (2, 0, -1, 0, 1, 16, 0, -8, 0),
    (-2, 2, 0, 2, 2, -16, 1, 7, 0),
    (0, 1, 0, 0, 1, -15, 0, 9, 0),
    (-2, 0, 1, 0, 1, -13, 0, 7, 0),
    (0, -1, 0, 0, 1, -12, 0, 6, 0),
    (0, 0, 2, -2, 0, 11, 0, 0, 0),
    (2, 0, -1, 2, 1, -10, 0, 5, 0),
    (2, 0, 1, 2, 2, -8, 0, 3, 0),
    (0, 1, 0, 2, 2, 7, 0, -3, 0),
    (-2, 1, 1, 0, 0, -7, 0, 0, 0),
    (0, -1, 0, 2, 2, -7, 0, 3, 0),
    (2, 0, 0, 2, 1, -7, 0, 3, 0),
    (2, 0, 1, 0, 0, 6, 0, 0, 0),
    (-2, 0, 2, 2, 2, 6, 0, -3, 0),
    (-2, 0, 1, 2, 1, 6, 0, -3, 0),
    (2, 0, -2, 0, 1, -6, 0, 3, 0),
    (2, 0, 0, 0, 1, -6, 0, 3, 0),
    (0, -1, 1, 0, 0, 5, 0, 0, 0),
    (-2, -1, 0, 2, 1, -5, 0, 3, 0),
    (-2, 0, 0, 0, 1, -5, 0, 3, 0),
    (0, 0, 2, 2, 1, -5, 0, 3, 0),
    (-2, 0, 2, 0, 1, 4, 0, 0, 0),
    (-2, 1, 0, 2, 1, 4, 0, 0, 0),
    (0, 0, 1, -2, 0, 4, 0, 0, 0),
    (-1, 0, 1, 0, 0, -4, 0, 0, 0),
    (-2, 1, 0, 0, 0, -4, 0, 0, 0),
    (1, 0, 0, 0, 0, -4, 0, 0, 0),
    (0, 0, 1, 2, 0, 3, 0, 0, 0),
    (0, 0, -2, 2, 2, -3, 0, 0, 0),
    (-1, -1, 1, 0, 0, -3, 0, 0, 0),
    (0, 1, 1, 0, 0, -3, 0, 0, 0),
    (0, -1, 1, 2, 2, -3, 0, 0, 0),
    (2, -1, -1, 2, 2, -3, 0, 0, 0),
    (0, 0, 3, 2, 2, -3, 0, 0, 0),
    (2, -1, 0, 2, 2, -3, 0, 0, 0),
)


def _constants(T):
    """Return some values needed for both nutation_in_longitude() and
    nutation_in_obliquity()"""
    D = modpi2(polynomial(kD, T))
    M = modpi2(polynomial(kM, T))
    M1 = modpi2(polynomial(kM1, T))
    F = modpi2(polynomial(kF, T))
    omega = modpi2(polynomial(ko, T))
    return D, M, M1, F, omega


def nutation_in_longitude(jd):
    """Return the nutation in longitude.

    High precision. [Meeus-1998: pg 144]

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - nutation in longitude, in radians

    """
    #
    # Future optimization: factor the /1e5 and /1e6 adjustments into the table.
    #
    # Could turn the loop into a generator expression. Too messy?
    #
    T = jd_to_jcent(jd)
    D, M, M1, F, omega = _constants(T)
    deltaPsi = 0.0
    for tD, tM, tM1, tF, tomega, tpsiK, tpsiT, tepsK, tepsT in _tbl:
        arg = D * tD + M * tM + M1 * tM1 + F * tF + omega * tomega
        deltaPsi += (tpsiK / 10000.0 + tpsiT / 100000.0 * T) * np.sin(arg)

    deltaPsi /= 3600
    deltaPsi = d_to_r(deltaPsi)
    return deltaPsi


def nutation_in_obliquity(jd):
    """Return the nutation in obliquity.

    High precision. [Meeus-1998: pg 144]

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - nutation in obliquity, in radians

    """
    #
    # Future optimization: factor the /1e5 and /1e6 adjustments into the table.
    #
    # Could turn the loop into a generator expression. Too messy?
    #
    T = jd_to_jcent(jd)
    D, M, M1, F, omega = _constants(T)
    deltaEps = 0.0
    for tD, tM, tM1, tF, tomega, tpsiK, tpsiT, tepsK, tepsT in _tbl:
        arg = D * tD + M * tM + M1 * tM1 + F * tF + omega * tomega
        deltaEps = deltaEps + (tepsK / 10000.0 + tepsT / 100000.0 * T) * np.cos(arg)
    deltaEps = deltaEps / 3600
    deltaEps = d_to_r(deltaEps)
    return deltaEps


#
# Constant terms
#
_el0 = (
    d_to_r(dms_to_d(23, 26, 21.448)),
    d_to_r(dms_to_d(0, 0, -46.8150)),
    d_to_r(dms_to_d(0, 0, -0.00059)),
    d_to_r(dms_to_d(0, 0, 0.001813)),
)


def obliquity(jd):
    """Return the mean obliquity of the ecliptic.

    Low precision, but good enough for most uses. [Meeus-1998: equation 22.2].
    Accuracy is 1" over 2000 years and 10" over 4000 years.

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - obliquity, in radians

    """
    T = jd_to_jcent(jd)
    return polynomial(_el0, T)


#
# Constant terms
#
_el1 = (
    d_to_r(dms_to_d(23, 26, 21.448)),
    d_to_r(dms_to_d(0, 0, -4680.93)),
    d_to_r(dms_to_d(0, 0, -1.55)),
    d_to_r(dms_to_d(0, 0, 1999.25)),
    d_to_r(dms_to_d(0, 0, -51.38)),
    d_to_r(dms_to_d(0, 0, -249.67)),
    d_to_r(dms_to_d(0, 0, -39.05)),
    d_to_r(dms_to_d(0, 0, 7.12)),
    d_to_r(dms_to_d(0, 0, 27.87)),
    d_to_r(dms_to_d(0, 0, 5.79)),
    d_to_r(dms_to_d(0, 0, 2.45)),
)


def obliquity_hi(jd):
    """Return the mean obliquity of the ecliptic.

    High precision [Meeus-1998: equation 22.3].

    Accuracy is 0.01" between 1000 and 3000, and "a few arc-seconds
    after 10,000 years".

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - obliquity, in radians

    """
    U = jd_to_jcent(jd) / 100
    return polynomial(_el1, U)



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

A collection of date and time functions.

The functions which use Julian Day Numbers are valid only for positive values,
i.e., for dates after -4712 (4713BC).

Unless otherwise specified, Julian Day Numbers may be fractional values.

Numeric years use the astronomical convention of a year 0: 0 = 1BC, -1 = 2BC,
etc.

Numeric months are 1-based: Jan = 1...Dec = 12.

Numeric days are the same as the calendar value.

Reference: Jean Meeus, _Astronomical Algorithms_, second edition, 1998,
Willmann-Bell, Inc.
"""

from math import modf

import numpy as np

from . import globals as globls
from .constants import minutes_per_day, seconds_per_day
from .util import _scalar_if_one, d_to_r, modpi2


class Error(Exception):
    """local exception class"""

    pass


def frac_yr_to_jd(year, gregorian=True):
    """Convert a date in the Julian or Gregorian fractional year to the
    Julian Day Number (Meeus 7.1).

    Arguments:
      - `year` : (int, float)  year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (float)

    """
    year = np.atleast_1d(year)
    day = np.atleast_1d(0.0).astype(np.float64)
    year, day = list(map(np.array, np.broadcast_arrays(year, day)))
    # For float years abuse the day variable
    fyear = year - year.astype("i")
    mask = fyear > 0
    if np.any(mask):
        year = year.astype("i")
        days_in_year = cal_to_jd(year[mask] + 1) - cal_to_jd(year[mask])
        day[mask] = days_in_year * fyear[mask]
        return _scalar_if_one(cal_to_jd(year) + day)
    return _scalar_if_one(cal_to_jd(year))


def yr_frac_mon_to_jd(year, mon, gregorian=True):
    """Convert a year and fractional month in the Julian or Gregorian
    calendars to the Julian Day Number (Meeus 7.1).

    Arguments:
      - `year` : (int)  year
      - `mon` : (int, float)  month

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (float)

    """
    year = np.atleast_1d(year)
    mon = np.atleast_1d(mon).astype(np.float64)
    day = np.atleast_1d(0.0).astype(np.float64)
    year, mon, day = list(map(np.array, np.broadcast_arrays(year, mon, day)))
    fmon = mon - mon.astype("i")
    mask = fmon > 0
    if np.any(mask):
        mon = mon.astype("i")
        next_mon = np.copy(mon) + 1
        next_year = np.copy(year)
        next_mon_mask = next_mon == 13
        next_year[next_mon_mask] = next_year[next_mon_mask] + 1
        next_mon[next_mon_mask] = 1
        days_in_mon = cal_to_jd(next_year[mask], next_mon[mask]) - cal_to_jd(
            year[mask], mon[mask]
        )
        day[mask] = days_in_mon * fmon[mask]
        return _scalar_if_one(cal_to_jd(year, mon) + day)
    return _scalar_if_one(cal_to_jd(year, mon))


def cal_to_jd(year, mon=1, day=1, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to the Julian Day
    Number (Meeus 7.1).

    Arguments:
      - `year` : (int)  year

    Keywords:
      - `mon`       : (int, default=1) month
      - `day`       : (int, float, default=1) day, may be fractional day
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (int, float)

    """
    year = np.atleast_1d(year)
    mon = np.atleast_1d(mon)
    day = np.atleast_1d(day).astype(np.float64)

    fyear = year - year.astype("i")
    mask = fyear > 0
    if np.any(mask):
        raise ValueError("Year must be integer. Use frac_yr_to_jd instead.")
    fmon = mon - mon.astype("i")
    mask = fmon > 0
    if np.any(mask):
        raise ValueError("Month must be integer. Use yr_frac_mon_to_jd instead.")
    if np.any(mon > 12) or np.any(mon < 1):
        raise ValueError("Month must be from 1 to 12")
    if np.any(day > 31) or np.any(day < 1):
        raise ValueError("Day must be from 1 to 31")
    year, mon, day = list(map(np.array, np.broadcast_arrays(year, mon, day)))

    for thirtydays in [9, 4, 6, 11]:
        daytestarr = mon == thirtydays
        if np.any(day[daytestarr] > 30):
            raise ValueError("Day must be from 1 to 30")

    leapyeartest = np.atleast_1d(is_leap_year(year, gregorian))

    if np.any(np.logical_and(day[leapyeartest] > 29, mon[leapyeartest] == 2)):
        raise ValueError("Day must be from 1 to 29")
    if np.any(np.logical_and(day[~leapyeartest] > 28, mon[~leapyeartest] == 2)):
        raise ValueError("Day must be from 1 to 28")

    testarr = mon <= 2
    year[testarr] -= 1
    mon[testarr] += 12
    if gregorian:
        A = (year / 100).astype(np.int64)
        B = 2 - A + (A / 4).astype(np.int64)
    else:
        B = 0
    return _scalar_if_one(
        (365.25 * (year + 4716)).astype(np.int64)
        + (30.6001 * (mon + 1)).astype(np.int64)
        + day
        + B
        - 1524.5
    )


def jd_to_cal(julian_day, gregorian=True):
    """Convert a Julian day number to a date in the Julian or Gregorian
    calendars.

    Arguments:
      - `julian_day` : (int) Julian Day Number

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - (year, month, day) : (tuple) day may be fractional

    """
    julian_day = np.atleast_1d(julian_day)
    F, Z = np.modf(julian_day + 0.5)
    if gregorian:
        alpha = ((Z - 1867216.25) / 36524.25).astype(np.int64)
        A = Z + 1 + alpha - (alpha / 4).astype(np.int64)
    else:
        A = Z
    B = A + 1524
    C = ((B - 122.1) / 365.25).astype(np.int64)
    D = (365.25 * C).astype(np.int64)
    E = ((B - D) / 30.6001).astype(np.int64)
    day = B - D - (30.6001 * E).astype(np.int64) + F
    mon = E - 13
    mon[E < 14] = E[E < 14] - 1
    year = C - 4715
    year[mon > 2] = C[mon > 2] - 4716
    return _scalar_if_one(year), _scalar_if_one(mon), _scalar_if_one(day)


def cal_to_jde(year, mon=1, day=1, hour=0, minute=0, sec=0.0, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to the Julian Day
    Ephemeris (Meeus 22.1).

    Arguments:
      - `year` : year

    Keywords:
      - `mon`       : (int, default=1) month
      - `day`       : (int, default=1) day, may be fractional day
      - `hour`      : (int, default=0) hour
      - `minute`    : (int, default=0) minute
      - `sec`       : (float, default=0.0) second
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - julian day ephemeris : (float)

    """
    jde = cal_to_jd(year, mon, day, gregorian)
    return _scalar_if_one(jde + hms_to_fday(hour, minute, sec))


def cal_to_day_of_year(year, mon, day, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to day of the year
    (Meeus 7.1).

    Arguments:
      - `year` : (int) year
      - `mon`  : (int) month
      - `day`  : (int) day

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - day number : 1 = Jan 1...365 (or 366 for leap years) = Dec 31.

    """
    year = np.atleast_1d(year).astype(np.int64)
    mon = np.atleast_1d(mon).astype(np.int64)
    day = np.atleast_1d(day).astype(np.int64)
    year, mon, day = np.broadcast_arrays(year, mon, day)
    K = np.ones_like(year)
    K[:] = 2
    K[np.atleast_1d(is_leap_year(year, gregorian))] = 1
    return _scalar_if_one(
        (275 * mon / 9.0).astype(np.int64)
        - (K * ((mon + 9) / 12.0).astype(np.int64))
        + day
        - 30
    )


def day_of_year_to_cal(year, N, gregorian=True):
    """Convert a day of year number to a month and day in the Julian or
    Gregorian calendars.

    Arguments:
      - `year`      : year
      - `N`         : day of year, 1..365 (or 366 for leap years)

    Keywords:
      - `gregorian` : If True, use Gregorian calendar, else use Julian calendar
        (default: True)

    Return:
      - (month, day) : (tuple)

    """
    year = np.atleast_1d(year)
    N = np.atleast_1d(N)
    year, N = np.broadcast_arrays(year, N)
    K = np.ones_like(N)
    K[:] = 2
    K[np.atleast_1d(is_leap_year(year, gregorian))] = 1
    mon = (9 * (K + N) / 275.0 + 0.98).astype(np.int64)
    mon[N < 32] = 1
    day = (
        N
        - (275 * mon / 9.0).astype(np.int64)
        + K * ((mon + 9) / 12.0).astype(np.int64)
        + 30
    ).astype(np.int64)
    return _scalar_if_one(mon), _scalar_if_one(day)


def easter(year, gregorian=True):
    """Return the date of Western ecclesiastical Easter for a year in the
    Julian or Gregorian calendars.

    Arguments:
      - `year` : (int) year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - (month, day) : (tuple)

    """
    year = np.atleast_1d(year)
    if gregorian:
        a = year % 19
        b = year // 100
        c = year % 100
        d = b // 4
        e = b % 4
        f = (b + 8) // 25
        g = (b - f + 1) // 3
        h = (19 * a + b - d - g + 15) % 30
        i = c // 4
        k = c % 4
        l = (32 + 2 * e + 2 * i - h - k) % 7
        m = (a + 11 * h + 22 * l) // 451
        tmp = h + l - 7 * m + 114
    else:
        a = year % 4
        b = year % 7
        c = year % 19
        d = (19 * c + 15) % 30
        e = (2 * a + 4 * b - d + 34) % 7
        tmp = d + e + 114
    mon = tmp // 31
    day = (tmp % 31) + 1
    return _scalar_if_one(mon), _scalar_if_one(day)


def fday_to_hms(day):
    """Convert fractional day (0.0..1.0) to integral hours, minutes, seconds.

    Arguments:
      - day : a fractional day in the range 0.0..1.0

    Returns:
      - hour : (int, 0..23)
      - minute : (int, 0..59)
      - second : (int, 0..59)

    """
    # First get rid of the integer day
    fday, days = modf(day)
    seconds = fday * 86400.0
    minutes = int(seconds / 60.0)
    seconds = seconds - (minutes * 60.0)
    hours = int(minutes / 60.0)
    minutes = minutes - (hours * 60.0)
    return int(hours), int(minutes), int(seconds)


def hms_to_fday(hr, mn, sec):
    """Convert hours-minutes-seconds into a fractional day 0.0..1.0.

    Arguments:
      - `hr` : hours, 0..23
      - `mn` : minutes, 0..59
      - `sec` : seconds, 0..59

    Returns:
      - fractional day, 0.0..1.0

    """
    hr = np.atleast_1d(hr)
    mn = np.atleast_1d(mn)
    sec = np.atleast_1d(sec)
    hr, mn, sec = np.broadcast_arrays(hr, mn, sec)
    return (hr / 24.0) + (mn / minutes_per_day) + (sec / seconds_per_day)


def is_dst(julian_day):
    """Is this instant within the Daylight Savings Time period.

    Uses the time zone database associated with Python and used in the 'time'
    module.

    Arguments:
      - `julian_day` : (int) Julian Day number representing an instant in
        Universal Time

    Returns:
      - (bool) True if Daylight Savings Time is in effect, False otherwise.

    """
    import datetime
    import time

    year, mon, day = jd_to_cal(julian_day)
    day = int(day)
    hr, minute, second = fday_to_hms(julian_day)
    second = int(second)

    stamp = datetime.datetime(year, mon, day, hr, minute, second)
    return time.localtime(time.mktime(stamp.timetuple())).tm_isdst == 1


def is_leap_year(year, gregorian=True):
    """Return True if this is a leap year in the Julian or Gregorian calendars

    Arguments:
      - `year` : (int) year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (bool) True is this is a leap year, else False.

    """
    year = np.atleast_1d(year).astype(np.int64)
    x = np.fmod(year, 4)
    if gregorian:
        x = np.fmod(year, 4)
        y = np.fmod(year, 100)
        z = np.fmod(year, 400)
        return _scalar_if_one(
            np.logical_and(np.logical_not(x), np.logical_or(y, np.logical_not(z)))
        )
    else:
        return _scalar_if_one(x == 0)


def jd_to_day_of_week(julian_day):
    """Return the day of week for a Julian Day Number.

    The Julian Day Number must be for 0h UT.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Returns:
      - day of week : (int) 0 = Sunday...6 = Saturday.

    """
    julian_day = np.atleast_1d(julian_day)
    i = (julian_day + 1.5).astype(np.int64)
    return _scalar_if_one(i % 7)


def jd_to_jcent(julian_day):
    """Return the number of Julian centuries since J2000.0

    Arguments:
      - `julian_day` : (int) Julian Day number

    Return:
      - Julian centuries : (int)

    """
    julian_day = np.atleast_1d(julian_day)
    return _scalar_if_one((julian_day - 2451545.0) / 36525.0)


def lt_to_str(julian_day, zone="", level="second"):
    """Convert local time in Julian Days to a formatted string.

    The general format is:

        YYYY-MMM-DD HH:MM:SS ZZZ

    Truncate the time value to seconds, minutes, hours or days as
    indicated. If level = "day", don't print the time zone string.

    Pass an empty string ("", the default) for zone if you want to do
    your own zone formatting in the calling module.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Keywords:
      - `zone`  : (str, default="") Time zone string
      - level : (str, default="second") {"day", "hour", "minute", "second"}

    Return:
      - formatted date/time string : (str)

    """
    year, mon, day = jd_to_cal(julian_day)
    fday, iday = modf(day)
    iday = int(iday)
    hour, minute, sec = fday_to_hms(fday)
    sec = int(sec)

    if level == "second":
        return "%d-%02d-%02d %02d:%02d:%02d %s" % (
            year,
            mon,
            iday,
            hour,
            minute,
            sec,
            zone,
        )
    if level == "minute":
        return "%d-%02d-%02d %02d:%02d %s" % (year, mon, iday, hour, minute, zone)
    if level == "hour":
        return "%d-%02d-%02d %02d %s" % (year, mon, iday, hour, zone)
    if level == "day":
        return "%d-%02d-%02d" % (year, mon, iday)

    raise Error("unknown time level = " + level)


def sidereal_time_greenwich(julian_day):
    """Return the mean sidereal time at Greenwich.

    The Julian Day number must represent Universal Time.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Returns:
      - sidereal time in radians : (float) 2pi radians = 24 hours

    """
    T = jd_to_jcent(julian_day)
    T2 = T * T
    T3 = T2 * T
    theta0 = (
        280.46061837
        + 360.98564736629 * (julian_day - 2451545.0)
        + 0.000387933 * T2
        - T3 / 38710000
    )
    result = d_to_r(theta0)
    return modpi2(result)


def ut_to_lt(julian_day):
    """Convert universal time in Julian Days to a local time.

    Include Daylight Savings Time offset, if any.

    Arguments:
      - `julian_day` : (int) Julian Day number, universal time

    Return:
      - Julian Day number : (str) local time
        zone string of the zone used for the conversion

    """
    if is_dst(julian_day):
        zone = globls.daylight_timezone_name
        offset = globls.daylight_timezone_offset
    else:
        zone = globls.standard_timezone_name
        offset = globls.standard_timezone_offset

    julian_day = julian_day - offset
    return julian_day, zone



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Compute Rise, Set, and Transit times.

Each of the routines requires three equatorial coordinates for the
object: yesterday, today and tomorrow, all at 0hr UT.

This approach is inadequate for the Moon, which moves too fast to
be accurately interpolated from three daily positions.

Bug: each of the routines drops some events which occur near 0hr UT.

"""

import numpy as np

from . import globals as globls
from .calendar import sidereal_time_greenwich
from .constants import earth_equ_radius, pi2, seconds_per_day, standard_rst_altitude
from .coordinates import equ_to_horiz
from .dynamical import deltaT_seconds
from .util import d_to_r, diff_angle, interpolate3, interpolate_angle3, modpi2


class Error(Exception):
    """local exception class"""

    pass


_k1 = d_to_r(360.985647)


def _riseset(
    jd,
    raList,
    decList,
    h0,
    delta,
    mode,
    longitude=globls.longitude,
    latitude=globls.latitude,
):
    # Private function since rise/set so similar

    THETA0 = sidereal_time_greenwich(jd)
    deltaT_days = deltaT_seconds(jd) / seconds_per_day

    cosH0 = (np.sin(h0) - np.sin(latitude) * np.sin(decList[1])) / (
        np.cos(latitude) * np.cos(decList[1])
    )
    #
    # future: return some indicator when the object is circumpolar or always
    # below the horizon.
    #
    if cosH0 < -1.0:  # circumpolar
        return None
    if cosH0 > 1.0:  # never rises
        return None

    H0 = np.arccos(cosH0)
    m0 = (raList[1] + longitude - THETA0) / pi2
    if mode == "rise":
        m = m0 - H0 / pi2  # the only difference between rise() and settime()
    elif mode == "set":
        m = m0 + H0 / pi2  # the only difference between rise() and settime()
    if m < 0:
        m += 1
    elif m > 1:
        m -= 1
    if not 0 <= m <= 1:
        raise Error("m is out of range = " + str(m))
    for bailout in range(20):
        m0 = m
        theta0 = modpi2(THETA0 + _k1 * m)
        n = m + deltaT_days
        if not -1 < n < 1:
            return None  # Bug: this is where we drop some events
        ra = interpolate_angle3(n, raList)
        dec = interpolate3(n, decList)
        H = theta0 - longitude - ra
        #        if H > pi:
        #            H = H - pi2
        H = diff_angle(0.0, H)
        A, h = equ_to_horiz(H, dec)
        dm = (h - h0) / (pi2 * np.cos(dec) * np.cos(latitude) * np.sin(H))
        m += dm
        if abs(m - m0) < delta:
            return jd + m

    raise Error("bailout")


def rise(jd, raList, decList, h0, delta):
    """Return the Julian Day of the rise time of an object.

    Arguments:
      - `jd`     : (int) Julian Day number of the day in question, at 0 hr UT
      - `raList` : (float, float, float) a sequence of three right accension
        values, in radians, for (jd-1, jd, jd+1)
      - `decList`: (float, float, float) a sequence of three right declination
        values, in radians, for (jd-1, jd, jd+1)
      - `h0`     : (float) the standard altitude in radians
      - `delta`  : (float) desired accuracy in days. Times less than one minute
        are infeasible for rise times because of atmospheric refraction.

    Returns:
      - Julian Day of the rise time

    """
    _riseset(jd, raList, decList, h0, delta, "rise")


def settime(jd, raList, decList, h0, delta):
    """Return the Julian Day of the set time of an object.

    Arguments:
      - `jd`      : Julian Day number of the day in question, at 0 hr UT
      - `raList`  : a sequence of three right accension values, in radians, for
        (jd-1, jd, jd+1)
      - `decList` : a sequence of three right declination values, in radians,
        for (jd-1, jd, jd+1)
      - `h0`      : the standard altitude in radians
      - `delta`   : desired accuracy in days. Times less than one minute are
        infeasible for set times because of atmospheric refraction.

    Returns:
      - Julian Day of the set time

    """
    _riseset(jd, raList, decList, h0, delta, "set")


def transit(jd, raList, delta):
    """Return the Julian Day of the transit time of an object.

    Arguments:
      - `jd`      : Julian Day number of the day in question, at 0 hr UT
      - `raList`  : a sequence of three right accension values, in radians, for
        (jd-1, jd, jd+1)
      - `delta`   : desired accuracy in days.

    Returns:
      - Julian Day of the transit time

    """
    #
    # future: report both upper and lower culmination, and transits of objects
    # below the horizon
    #
    longitude = globls.longitude
    THETA0 = sidereal_time_greenwich(jd)
    deltaT_days = deltaT_seconds(jd) / seconds_per_day

    m = (raList[1] + longitude - THETA0) / pi2
    if m < 0:
        m += 1
    elif m > 1:
        m -= 1
    if not 0 <= m <= 1:
        raise Error("m is out of range = " + str(m))
    for bailout in range(20):
        m0 = m
        theta0 = modpi2(THETA0 + _k1 * m)
        n = m + deltaT_days
        if not -1 < n < 1:
            return None  # Bug: this is where we drop some events
        ra = interpolate_angle3(n, raList)
        H = theta0 - longitude - ra
        #        if H > pi:
        #            H = H - pi2
        H = diff_angle(0.0, H)
        dm = -H / pi2
        m += dm
        if abs(m - m0) < delta:
            return jd + m

    raise Error("bailout")


def moon_rst_altitude(r):
    """Return the standard altitude of the Moon.

    Arguments:
      - `r` : Distance between the centers of the Earth and Moon, in km.

    Returns:
      - Standard altitude in radians.

    """
    # horizontal parallax
    parallax = np.arcsin(earth_equ_radius / r)

    return 0.7275 * parallax + standard_rst_altitude



# -*- coding: utf-8 -*-
from unittest import TestCase

from astronomia.coordinates import ecl_to_equ, ell_to_geo, equ_to_ecl
from astronomia.util import d_to_r, modpi2, r_to_d


class TestCoords(TestCase):
    def test_equ_to_ecl(self):
        longitude, latitude = equ_to_ecl(
            d_to_r(116.328942), d_to_r(28.026183), d_to_r(23.4392911)
        )

        self.assertAlmostEqual(r_to_d(longitude), 113.215630, places=6)
        self.assertAlmostEqual(r_to_d(latitude), 6.684170, places=6)

    def test_ecl_to_equ(self):
        ra, dec = ecl_to_equ(d_to_r(113.215630), d_to_r(6.684170), d_to_r(23.4392911))
        self.assertAlmostEqual(r_to_d(ra), 116.328942, places=5)
        self.assertAlmostEqual(r_to_d(dec), 28.026183, places=6)

    def test_ell_to_geo(self):
        phi, theta, r = ell_to_geo(d_to_r(0), d_to_r(0), 10000)
        self.assertAlmostEqual(r_to_d(modpi2(phi)), 203.23542197)
        self.assertAlmostEqual(r_to_d(modpi2(theta)), 90.0)
        self.assertAlmostEqual(r, 0.0)



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

from unittest import TestCase

from astronomia.nutation import (
    nutation_in_longitude,
    nutation_in_obliquity,
    obliquity,
    obliquity_hi,
)
from astronomia.util import d_to_dms, r_to_d


class TestNutation(TestCase):
    def test_nut_in_lon(self):
        deltaPsi = nutation_in_longitude(2446895.5)
        d, m, s = d_to_dms(r_to_d(deltaPsi))
        self.assertEqual(d, 0)
        self.assertEqual(m, 0)
        self.assertAlmostEqual(s, -3.788, places=3)

        # from http://www.neoprogrammics.com/nutations/Nutation_In_Longitude_And_RA.php
        deltaPsi = nutation_in_longitude(2456479.5)
        d, m, s = d_to_dms(r_to_d(deltaPsi))
        self.assertEqual(d, 0)
        self.assertEqual(m, 0)
        self.assertAlmostEqual(s, 12.675, places=3)

    def test_nut_in_obl(self):
        deltaEps = nutation_in_obliquity(2446895.5)
        d, m, s = d_to_dms(r_to_d(deltaEps))
        self.assertEqual(d, 0)
        self.assertEqual(m, 0)
        self.assertAlmostEqual(s, 9.443, places=3)

    def test_obliquity(self):
        eps = obliquity(2446895.5)
        d, m, s = d_to_dms(r_to_d(eps))
        self.assertEqual(d, 23)
        self.assertEqual(m, 26)
        self.assertAlmostEqual(s, 27.407, places=3)

    def test_obliquity_hi(self):
        eps = obliquity_hi(2446895.5)
        d, m, s = d_to_dms(r_to_d(eps))
        self.assertEqual(d, 23)
        self.assertEqual(m, 26)
        self.assertAlmostEqual(s, 27.407, places=3)



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Collection of miscellaneous functions
    """

import os
import shlex
import sys

import numpy as np

from . import globals as globls
from .constants import minutes_per_day, pi2, seconds_per_day


class Error(Exception):
    """Local exception class"""

    pass


def _scalar_if_one(solution):
    """Returns a scalar if array size is 1"""
    if np.isscalar(solution):
        return solution
    if solution.size == 1:
        return solution.item()
    return solution


def d_to_dms(x):
    """Convert an angle in decimal degrees to degree components.

    Return a tuple (degrees, minutes, seconds). Degrees and minutes
    will be integers, seconds may be floating.

    If the argument is negative:

        The return value of degrees will be negative.
        If degrees is 0, minutes will be negative.
        If minutes is 0, seconds will be negative.

    Arguments:
      - `x` : degrees

    Returns:
      - degrees : (int)
      - minutes : (int)
      - seconds : (int, float)

    """
    frac, degrees = np.modf(x)
    seconds, minutes = np.modf(frac * 60)
    return int(degrees), int(minutes), seconds * 60


#
# Local constants
#
_DtoR = np.pi / 180.0


def d_to_r(d):
    """Convert degrees to radians.

    Arguments:
      -  `d` : (int, float), degrees

    Returns:
      - radians : (float)

    """
    return d * _DtoR


def diff_angle(a, b):
    """Return angle b - a, accounting for circular values.

    Parameters a and b should be in the range 0..pi*2. The
    result will be in the range -pi..pi.

    This allows us to directly compare angles which cross through 0:

        359 degress... 0 degrees... 1 degree... etc

    Arguments:
      - `a` : (int, float) first angle, in radians
      - `b` : (int, float) second angle, in radians

    Returns:
      - b - a, in radians : (int, float)

    """
    if b < a:
        result = b + pi2 - a
    else:
        result = b - a
    if result > np.pi:
        result -= pi2
    return result


def dms_to_d(deg, minute, sec):
    """Convert an angle in degree components to decimal degrees.

    If any of the components are negative the result will also be negative.

    Arguments:
      - `deg` : (int, float) degrees
      - `minute` : (int, float) minutes
      - `sec` : (int, float) seconds

    Returns:
      - decimal degrees : (float)

    """
    deg = np.atleast_1d(deg)
    minute = np.atleast_1d(minute)
    sec = np.atleast_1d(sec)
    deg, minute, sec = np.broadcast_arrays(deg, minute, sec)
    result = abs(deg) + abs(minute) / 60.0 + abs(sec) / 3600.0
    if deg < 0 or minute < 0 or sec < 0:
        result = -result
    return _scalar_if_one(result)


def interpolate3(n, y):
    """Interpolate from three equally spaced tabular values.

    [Meeus-1998; equation 3.3]

    Parameters:
      - `n` : the interpolating factor, must be between -1 and 1
      - `y` : a sequence of three values

    Results:
      - the interpolated value of y

    """
    if not -1 < n < 1:
        raise Error("interpolating factor out of range: " + str(n))

    a = y[1] - y[0]
    b = y[2] - y[1]
    c = b - a
    return y[1] + n / 2 * (a + b + n * c)


def interpolate_angle3(n, y):
    """Interpolate from three equally spaced tabular angular values.

    [Meeus-1998; equation 3.3]

    This version is suitable for interpolating from a table of
    angular values which may cross the origin of the circle,
    for example: 359 degrees...0 degrees...1 degree.

    Arguments:
      - `n` : the interpolating factor, must be between -1 and 1
      - `y` : a sequence of three values

    Results:
      - the interpolated value of y

    """
    if not -1 < n < 1:
        raise Error("interpolating factor out of range: " + str(n))

    a = diff_angle(y[0], y[1])
    b = diff_angle(y[1], y[2])
    c = diff_angle(a, b)
    return y[1] + n / 2 * (a + b + n * c)


def load_params():
    """Read a parameter file and assign global values.

    Arguments:
        none

    Returns:
        nothing

    """
    fname = os.environ.get("ASTRONOMIA_PARAMS", "astronomia_params.txt")

    if not os.path.exists(fname):
        # last resort
        fname = os.path.join(sys.prefix, "share", "astronomia", "astronomia_params.txt")
        print(
            """WARNING: Using system wide settings file at
"{0}".
You may want to set the ASTRONOMIA_PARAMS environment variable to point to the
file you want, or create a "astronomia_params.txt" file in the current
directory.""".format(
                fname
            )
        )
    try:
        f = open(fname, "r")
    except IOError as value:
        raise Error(
            """
Unable to open param file. Either set ASTRONOMIA_PARAMS correctly or create
astronomia_params.txt in the current directory"""
        )

    lex = shlex.shlex(f)
    # tokens and values can have dots, dashes, slashes, colons
    lex.wordchars = lex.wordchars + ".-/\\:"
    token = lex.get_token()
    while token:
        if token == "standard_timezone_name":
            globls.standard_timezone_name = lex.get_token()
        elif token == "standard_timezone_offset":
            offset = float(lex.get_token())
            unit = lex.get_token().lower()
            if unit not in (
                "day",
                "days",
                "hour",
                "hours",
                "minute",
                "minutes",
                "second",
                "seconds",
            ):
                raise Error("bad value for standard_timezone_offset units")
            if unit in ("hour", "hours"):
                offset /= 24.0
            elif unit in ("minute", "minutes"):
                offset /= minutes_per_day
            elif unit in ("second", "seconds"):
                offset /= seconds_per_day
            globls.standard_timezone_offset = offset
        elif token == "daylight_timezone_name":
            globls.daylight_timezone_name = lex.get_token()
        elif token == "daylight_timezone_offset":
            offset = float(lex.get_token())
            unit = lex.get_token().lower()
            if unit not in (
                "day",
                "days",
                "hour",
                "hours",
                "minute",
                "minutes",
                "second",
                "seconds",
            ):
                raise Error("bad value for standard_timezone_offset units")
            if unit in ("hour", "hours"):
                offset /= 24.0
            elif unit in ("minute", "minutes"):
                offset /= minutes_per_day
            elif unit in ("second", "seconds"):
                offset /= seconds_per_day
            globls.daylight_timezone_offset = offset
        elif token == "longitude":
            longitude = float(lex.get_token())
            direction = lex.get_token().lower()
            if direction not in ("east", "west"):
                raise Error('longitude direction must be "west" or "east"')
            if direction == "east":
                longitude = -longitude
            globls.longitude = d_to_r(longitude)
        elif token == "latitude":
            latitude = float(lex.get_token())
            direction = lex.get_token().lower()
            if direction not in ("north", "south"):
                raise Error('latitude direction must be "north" or "south"')
            if direction == "south":
                latitude = -latitude
            globls.latitude = d_to_r(latitude)
        elif token == "vsop87d_text_path":
            globls.vsop87d_text_path = lex.get_token()
        elif token == "vsop87d_binary_path":
            globls.vsop87d_binary_path = lex.get_token()
        else:
            raise Error(
                "unknown token %s at line %d in param file" % (token, lex.lineno)
            )
        token = lex.get_token()

    f.close()


def modpi2(x):
    """Reduce an angle in radians to the range 0..2pi.

    Arguments:
      - `x` : angle in radians

    Returns:
      - angle in radians in the range 0..2pi

    """
    return x % pi2


def mod360(x):
    """Reduce an angle in degrees to the range 0..360.

    Arguments:
      - `x` : angle in degrees

    Returns:
      - angle in degress in the range 0..360

    """
    return x % 360


def polynomial(terms, x):
    """Evaluate a simple polynomial.

    Where: terms[0] is constant, terms[1] is for x, terms[2] is for x^2, etc.

    Arguments:
      - `terms` : sequence of coefficients
      - `x` : variable value

    Results:
      - value of the polynomial

    Examples:
        >>> t = 4.1
        >>> polynomial((1.1, -3.2, 3.3, 4.5), t)
        353.59749999999997

        returns the value of:

            1.1 + 2.2 * t + 3.3 * t^2 + 4.4 * t^3
    """
    apolyfunc = np.polynomial.Polynomial(terms)
    return apolyfunc(x)


#
# Local constants
#
_RtoD = 180.0 / np.pi


def r_to_d(r):
    """Convert radians to degrees.

    Arguments:
      - `r` : radians

    Returns:
      - degrees

    """
    return r * _RtoD



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

A throw-away script. How fast does the radius vector change
around the time of summer solstice?

Usage:

    ./check_perihelion.py

"""

from astronomia.calendar import cal_to_jd
from astronomia.constants import km_per_au
from astronomia.planets import VSOP87d
from astronomia.util import load_params

load_params()
vsop = VSOP87d()

exact_jd = cal_to_jd(1991, 7, 6 + (15.46 / 24))

exact_R = vsop.dimension(exact_jd, "Earth", "R")

days_per_hour = 1.0 / 24

for i in range(-24, 24):
    jd = exact_jd + i * days_per_hour
    R = vsop.dimension(jd, "Earth", "R")
    print(i, (exact_R - R) * km_per_au)



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Calculate the times of solstice and equinox events for Earth
    """

from math import cos, pi

from . import globals as globls
from .calendar import jd_to_jcent
from .constants import pi2
from .nutation import nutation_in_longitude
from .planets import vsop_to_fk5
from .sun import Sun, aberration_low
from .util import d_to_r, diff_angle, polynomial


class Error(Exception):
    """local exception class"""

    pass


#
# Meeus-1998 Table 27.A
#
_approx_1000 = {
    "spring": (1721139.29189, 365242.13740, 0.06134, 0.00111, -0.00071),
    "summer": (1721233.25401, 365241.72562, -0.05323, 0.00907, -0.00025),
    "autumn": (1721325.70455, 365242.49558, -0.11677, -0.00297, 0.00074),
    "winter": (1721414.39987, 365242.88257, -0.00769, -0.00933, -0.00006),
}

#
# Meeus-1998 Table 27.B
#
_approx_3000 = {
    "spring": (2451623.80984, 365242.37404, 0.05169, -0.00411, -0.00057),
    "summer": (2451716.56767, 365241.62603, 0.00325, 0.00888, -0.00030),
    "autumn": (2451810.21715, 365242.01767, -0.11575, 0.00337, 0.00078),
    "winter": (2451900.05952, 365242.74049, -0.06223, -0.00823, 0.00032),
}

#
# Meeus-1998 Table 27.C
#
_terms = [
    (485, d_to_r(324.96), d_to_r(1934.136)),
    (203, d_to_r(337.23), d_to_r(32964.467)),
    (199, d_to_r(342.08), d_to_r(20.186)),
    (182, d_to_r(27.85), d_to_r(445267.112)),
    (156, d_to_r(73.14), d_to_r(45036.886)),
    (136, d_to_r(171.52), d_to_r(22518.443)),
    (77, d_to_r(222.54), d_to_r(65928.934)),
    (74, d_to_r(296.72), d_to_r(3034.906)),
    (70, d_to_r(243.58), d_to_r(9037.513)),
    (58, d_to_r(119.81), d_to_r(33718.147)),
    (52, d_to_r(297.17), d_to_r(150.678)),
    (50, d_to_r(21.02), d_to_r(2281.226)),
    (45, d_to_r(247.54), d_to_r(29929.562)),
    (44, d_to_r(325.15), d_to_r(31555.956)),
    (29, d_to_r(60.93), d_to_r(4443.417)),
    (18, d_to_r(155.12), d_to_r(67555.328)),
    (17, d_to_r(288.79), d_to_r(4562.452)),
    (16, d_to_r(198.04), d_to_r(62894.029)),
    (14, d_to_r(199.76), d_to_r(31436.921)),
    (12, d_to_r(95.39), d_to_r(14577.848)),
    (12, d_to_r(287.11), d_to_r(31931.756)),
    (12, d_to_r(320.81), d_to_r(34777.259)),
    (9, d_to_r(227.73), d_to_r(1222.114)),
    (8, d_to_r(15.45), d_to_r(16859.074)),
]


def equinox_approx(yr, season):
    """Returns the approximate time of a solstice or equinox event.

    The year must be in the range -1000...3000. Within that range the
    the error from the precise instant is at most 2.16 minutes.

    Arguments:
      - `yr`     : (int) year
      - `season` : (str) {"spring", "summer", "autumn", "winter"}

    Returns:
      - Julian Day : (int) in dynamical time

    """
    if not (-1000 <= yr <= 3000):
        raise Error("year is out of range")
    if season not in globls.season_names:
        raise Error("unknown season =" + season)

    yr = int(yr)
    if -1000 <= yr <= 1000:
        Y = yr / 1000.0
        tbl = _approx_1000
    else:
        Y = (yr - 2000) / 1000.0
        tbl = _approx_3000

    jd = polynomial(tbl[season], Y)
    T = jd_to_jcent(jd)
    W = d_to_r(35999.373 * T - 2.47)
    delta_lambda = 1 + 0.0334 * cos(W) + 0.0007 * cos(2 * W)

    jd += 0.00001 * sum([A * cos(B + C * T) for A, B, C in _terms]) / delta_lambda

    return jd


_circle = {"spring": 0.0, "summer": pi * 0.5, "autumn": pi, "winter": pi * 1.5}

_k_sun_motion = 365.25 / pi2


def equinox(jd, season, delta):
    """Return the precise moment of an equinox or solstice event on Earth.

    Parameters:
      - `jd`     : Julian of an approximate time of the event in dynamical time
      - `season` : one of ("spring", "summer", "autumn", "winter")
      - `delta`  : the required precision in days. Times accurate to a second
        are reasonable when using the VSOP model.

    Returns:
      - Julian Day : (int) dynamical time

    """
    #
    # If we knew that the starting approximate time was close enough
    # to the actual time, we could pull nutation_in_longitude() and the
    # aberration out of the loop and save some calculating.
    #
    circ = _circle[season]
    sun = Sun()
    for i in range(20):
        jd0 = jd
        L, B, R = sun.dimension3(jd)
        L += nutation_in_longitude(jd) + aberration_low(R)
        L, B = vsop_to_fk5(jd, L, B)
        # Meeus uses jd + 58 * sin(diff(...))
        jd += diff_angle(L, circ) * _k_sun_motion
        if abs(jd - jd0) < delta:
            return jd
    raise Error("bailout")



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The VSOP87d planetary position model
    """

import numpy as np

from .calendar import jd_to_jcent
from .constants import pi2
from .coordinates import ecl_to_equ
from .util import _scalar_if_one, d_to_r, diff_angle, dms_to_d, modpi2, polynomial


class Error(Exception):
    """Local exception class"""

    pass


#
# Global values, readable from other modules
#
planet_names = (
    "Mercury",
    "Venus",
    "Earth",
    "Mars",
    "Jupiter",
    "Saturn",
    "Uranus",
    "Neptune",
)
coordinate_names = ("L", "B", "R")

#
# Local dictionary of planetary terms.
#
# The key is a tuple (planet_name, coordinate_name)
#
# The value of each entry is a list of lists.
#
_planets = {}

_first_time = True


class VSOP87d:
    """The VSOP87d planetary model.

    All instances of this class share a common dictionary
    of planetary terms.

    """

    def __init__(self):
        """Load the database of planetary terms. This is actually done
        only once to save time and space.

        """
        global _first_time

        if not _first_time:
            return

        load_vsop87d_text_db()
        _first_time = False

    def dimension(self, jd, planet, dim):
        """Return one of heliocentric ecliptic longitude, latitude and radius.
        [Meeus-1998: pg 218]

        Arguments:
          - `jd`     : Julian Day in dynamical time
          - `planet` : must be one of ("Mercury", "Venus", "Earth", "Mars",
            "Jupiter", "Saturn", "Uranus", "Neptune")
          - `dim`    : must be one of "L" (longitude) or "B" (latitude) or "R"
            (radius)

        Returns:
          - longitude in radians, or latitude in radians, or radius in au,
            depending on the value of `dim`.

        """
        jd = np.atleast_1d(jd)
        X = 0.0
        tauN = 1.0
        tau = jd_to_jcent(jd) / 10.0
        c = _planets[(planet, dim)]

        for s in c:
            X += np.sum([A * np.cos(B + C * tau) for A, B, C in s]) * tauN
            tauN = tauN * tau  # last calculation is wasted

        if dim == "L":
            X = modpi2(X)

        return _scalar_if_one(X)

    def dimension3(self, jd, planet):
        """Return heliocentric ecliptic longitude, latitude and radius.

        Arguments:
          - `jd`     : Julian Day in dynamical time
          - `planet` : must be one of ("Mercury", "Venus", "Earth", "Mars",
            "Jupiter", "Saturn", "Uranus", "Neptune")

        Returns:
          - longitude in radians
          - latitude in radians
          - radius in au

        """
        L = self.dimension(jd, planet, "L")
        B = self.dimension(jd, planet, "B")
        R = self.dimension(jd, planet, "R")
        return L, B, R


#
# Constant terms
#
_k0 = d_to_r(-1.397)
_k1 = d_to_r(-0.00031)
_k2 = d_to_r(dms_to_d(0, 0, -0.09033))
_k3 = d_to_r(dms_to_d(0, 0, 0.03916))


def vsop_to_fk5(jd, L, B):
    """Convert VSOP to FK5 coordinates.

    This is required only when using the full precision of the
    VSOP model.  [Meeus-1998: pg 219]

    Arguments:
      - `jd` : Julian Day in dynamical time
      - `L`  : longitude in radians
      - `B`  : latitude in radians

    Returns:
      - corrected longitude in radians
      - corrected latitude in radians

    """
    jd = np.atleast_1d(jd)
    T = jd_to_jcent(jd)
    L1 = polynomial([L, _k0, _k1], T)
    cosL1 = np.cos(L1)
    sinL1 = np.sin(L1)
    deltaL = _k2 + _k3 * (cosL1 + sinL1) * np.tan(B)
    deltaB = _k3 * (cosL1 - sinL1)
    return _scalar_if_one(modpi2(L + deltaL)), _scalar_if_one(B + deltaB)


def geocentric_planet(jd, planet, deltaPsi, epsilon, delta):
    """Calculate the equatorial coordinates of a planet

    The results will be geocentric, corrected for light-time and
    aberration.

    Arguments:
      - `jd`       : Julian Day in dynamical time
      - `planet`   : must be one of ("Mercury", "Venus", "Earth", "Mars",
        "Jupiter", "Saturn", "Uranus", "Neptune")
      - `deltaPsi` : nutation in longitude, in radians
      - `epsilon`  : True obliquity (corrected for nutation), in radians
      - `delta`    : desired accuracy, in days

    Returns:
      - right accension, in radians
      - declination, in radians

    """
    jd = np.atleast_1d(jd)
    vsop = VSOP87d()
    t = jd
    l0 = -100.0  # impossible value
    # We need to iterate to correct for light-time and aberration.
    # At most three passes through the loop always nails it.
    # Note that we move both the Earth and the other planet during
    #    the iteration.
    for bailout in range(20):
        # heliocentric geometric ecliptic coordinates of the Earth
        L0, B0, R0 = vsop.dimension3(t, "Earth")

        # heliocentric geometric ecliptic coordinates of the planet
        L, B, R = vsop.dimension3(t, planet)

        # rectangular offset
        cosB0 = np.cos(B0)
        cosB = np.cos(B)
        x = R * cosB * np.cos(L) - R0 * cosB0 * np.cos(L0)
        y = R * cosB * np.sin(L) - R0 * cosB0 * np.sin(L0)
        z = R * np.sin(B) - R0 * np.sin(B0)

        # geocentric geometric ecliptic coordinates of the planet
        x2 = x * x
        y2 = y * y
        l = np.arctan2(y, x)
        b = np.arctan2(z, np.sqrt(x2 + y2))

        # distance to planet in AU
        dist = np.sqrt(x2 + y2 + z * z)

        # light time in days
        tau = 0.0057755183 * dist

        if abs(diff_angle(l, l0)) < pi2 * delta:
            break

        # adjust for light travel time and try again
        l0 = l
        t = jd - tau
    else:
        raise Error("bailout")

    # transform to FK5 ecliptic and equinox
    l, b = vsop_to_fk5(jd, l, b)

    # nutation in longitude
    l = l + deltaPsi

    # equatorial coordinates
    ra, dec = ecl_to_equ(l, b, epsilon)

    return ra, dec


def load_vsop87d_text_db():
    """Load the text version of the VSOP87d database into memory.

    IMPORTANT: normally you don't call this routine directly.
    That is done automatically by the __init__() method of the VSOP87d
    class.

    """
    global _planets
    from .vsop87d_dict import _planets



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
A clock application that displays a variety of celestial events in the
order they occur.

Usage:

    ./cronus.py start_year [stop_year]

To do:
    -- Add many more events
    -- Support both real-time and "fast" modes
    -- Allow finer start and stop times

Currently the program always runs in "fast" mode, queueing and
displaying events in the future as fast as possible. Eventually
I would like to have enough events covered so that the display
runs continuously even in real-time. Since the next event of
a given type needs to be calculated only when the previous one
has been delivered, this is not as computationally intense as it
sounds.

"""

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import os
import sys
from heapq import heappop, heappush
from math import *

import astronomia.globals
from astronomia.calendar import cal_to_jd, easter, lt_to_str, ut_to_lt
from astronomia.constants import (
    days_per_minute,
    days_per_second,
    standard_rst_altitude,
    sun_rst_altitude,
)
from astronomia.coordinates import ecl_to_equ
from astronomia.dynamical import dt_to_ut
from astronomia.equinox import equinox, equinox_approx
from astronomia.lunar import Lunar
from astronomia.nutation import nutation_in_longitude, nutation_in_obliquity, obliquity
from astronomia.planets import VSOP87d, geocentric_planet, planet_names, vsop_to_fk5
from astronomia.riseset import moon_rst_altitude, rise, settime, transit
from astronomia.sun import Sun, aberration_low
from astronomia.util import load_params

vsop = None  # delay loading this until we are sure the script can run
sun = None  # "                  "
moon = Lunar()

HIGH_PRIORITY = 0.0

rstDict = {}


class Task:
    def __init__(self, jd, func, args):
        self.jd = jd
        self.func = func
        self.args = args

    def __cmp__(self, other):
        return cmp(self.jd, other.jd)

    def __gt__(self, other):
        return self.jd > other.jd


taskQueue = []


class RiseSetTransit:
    def __init__(self, name, raList, decList, h0List):
        self.name = name
        self.raList = raList
        self.decList = decList
        self.h0List = h0List


def display(str):
    print(str)


def doEaster(year):
    month, day = easter(year)
    jd = cal_to_jd(year, month, day)
    str = "%-24s %s" % (lt_to_str(jd, None, "day"), "Easter")
    heappush(taskQueue, Task(jd, display, (str,)))
    # recalculate on March 1, next year
    heappush(taskQueue, Task(cal_to_jd(year + 1, 3, 1), doEaster, (year + 1,)))


_seasons = {
    "spring": "Vernal Equinox",
    "summer": "Summer Solstice",
    "autumn": "Autumnal Equinox",
    "winter": "Winter Solstice",
}


def doEquinox(year, season):
    approx_jd = equinox_approx(year, season)
    jd = equinox(approx_jd, season, days_per_second)
    ut = dt_to_ut(jd)
    lt, zone = ut_to_lt(ut)
    str = lt_to_str(lt, zone) + " " + _seasons[season]
    heappush(taskQueue, Task(jd, display, (str,)))
    heappush(taskQueue, Task(jd, doEquinox, (year + 1, season)))


def doRiseSetTransit(jd_today):
    #
    # Find and queue rise-set-transit times for all objects
    #
    jd = jd_today
    for obj in list(rstDict.values()):
        td = rise(jd, obj.raList, obj.decList, obj.h0List[1], days_per_minute)
        if td:
            ut = dt_to_ut(td)
            lt, zone = ut_to_lt(ut)
            str = "%-19s %s %s rises" % (lt_to_str(lt, "", "minute"), zone, obj.name)
            heappush(taskQueue, Task(td, display, (str,)))
        else:
            print("****** RiseSetTransit failure:", obj.name, "rise")

        td = settime(jd, obj.raList, obj.decList, obj.h0List[1], days_per_minute)
        if td:
            ut = dt_to_ut(td)
            lt, zone = ut_to_lt(ut)
            str = "%-19s %s %s sets" % (lt_to_str(lt, "", "minute"), zone, obj.name)
            heappush(taskQueue, Task(td, display, (str,)))
        else:
            print("****** RiseSetTransit failure:", obj.name, "set")

        td = transit(jd, obj.raList, days_per_second)
        if td:
            ut = dt_to_ut(td)
            lt, zone = ut_to_lt(ut)
            str = "%-23s %s transits" % (lt_to_str(lt, zone), obj.name)
            heappush(taskQueue, Task(td, display, (str,)))
        else:
            print("****** RiseSetTransit failure:", obj.name, "transit")

    #
    # setup the day after tomorrow
    #
    jd += 2

    # nutation in longitude
    deltaPsi = nutation_in_longitude(jd)

    # apparent obliquity
    eps = obliquity(jd) + nutation_in_obliquity(jd)

    #
    # Planets
    #
    for planet in planet_names:
        if planet == "Earth":
            continue
        ra, dec = geocentric_planet(jd, planet, deltaPsi, eps, days_per_second)
        obj = rstDict[planet]
        del obj.raList[0]
        del obj.decList[0]
        del obj.h0List[0]
        obj.raList.append(ra)
        obj.decList.append(dec)
        obj.h0List.append(standard_rst_altitude)
    #
    # Moon
    #
    l, b, r = moon.dimension3(jd)

    # nutation in longitude
    l += deltaPsi

    # equatorial coordinates
    ra, dec = ecl_to_equ(l, b, eps)

    obj = rstDict["Moon"]
    del obj.raList[0]
    del obj.decList[0]
    del obj.h0List[0]
    obj.raList.append(ra)
    obj.decList.append(dec)
    obj.h0List.append(moon_rst_altitude(r))

    #
    # Sun
    #
    l, b, r = sun.dimension3(jd)

    # correct vsop coordinates
    l, b = vsop_to_fk5(jd, l, b)

    # nutation in longitude
    l += deltaPsi

    # aberration
    l += aberration_low(r)

    # equatorial coordinates
    ra, dec = ecl_to_equ(l, b, eps)

    obj = rstDict["Sun"]
    del obj.raList[0]
    del obj.decList[0]
    del obj.h0List[0]
    obj.raList.append(ra)
    obj.decList.append(dec)
    obj.h0List.append(sun_rst_altitude)

    heappush(taskQueue, Task(jd, doRiseSetTransit, (jd_today + 1,)))


def initRST(start_year):
    start_jd = cal_to_jd(start_year)

    #
    # We need nutation values for each of three days
    #
    nutation = {}
    for day in (-1, 0, 1):
        jd = start_jd + day
        # nutation in longitude
        deltaPsi = nutation_in_longitude(jd)
        # apparent obliquity
        eps = obliquity(jd) + nutation_in_obliquity(jd)
        nutation[day] = deltaPsi, eps

    #
    # Planets
    #
    for planet in planet_names:
        if planet == "Earth":
            continue
        raList = []
        decList = []
        h0List = []
        for day in (-1, 0, 1):
            jd = start_jd + day
            deltaPsi, eps = nutation[day]
            ra, dec = geocentric_planet(jd, planet, deltaPsi, eps, days_per_second)
            raList.append(ra)
            decList.append(dec)
            h0List.append(standard_rst_altitude)
        rstDict[planet] = RiseSetTransit(planet, raList, decList, h0List)

    #
    # Moon
    #
    raList = []
    decList = []
    h0List = []
    for day in (-1, 0, 1):
        jd = start_jd + day
        deltaPsi, eps = nutation[day]
        l, b, r = moon.dimension3(jd)
        # nutation in longitude
        l += deltaPsi
        # equatorial coordinates
        ra, dec = ecl_to_equ(l, b, eps)
        raList.append(ra)
        decList.append(dec)
        h0List.append(moon_rst_altitude(r))
    rstDict["Moon"] = RiseSetTransit("Moon", raList, decList, h0List)

    #
    # Sun
    #
    raList = []
    decList = []
    h0List = []
    for day in (-1, 0, 1):
        jd = start_jd + day
        deltaPsi, eps = nutation[day]
        l, b, r = sun.dimension3(jd)
        # correct vsop coordinates
        l, b = vsop_to_fk5(jd, l, b)
        # nutation in longitude
        l += deltaPsi
        # aberration
        l += aberration_low(r)
        # equatorial coordinates
        ra, dec = ecl_to_equ(l, b, eps)
        raList.append(ra)
        decList.append(dec)
        h0List.append(sun_rst_altitude)
    rstDict["Sun"] = RiseSetTransit("Sun", raList, decList, h0List)

    # all Rise-Set-Transit events
    heappush(taskQueue, Task(HIGH_PRIORITY, doRiseSetTransit, (start_jd,)))


def run():
    global vsop
    global sun
    if len(sys.argv) < 2:
        print(__doc__)
        os._exit(0)
    if len(sys.argv) < 3:
        start_year = int(sys.argv[1])
        stop_jd = cal_to_jd(10000)  # default stopping date: 10,000AD
    elif len(sys.argv) < 4:
        start_year = int(sys.argv[1])
        stop_jd = cal_to_jd(int(sys.argv[2]))
    else:
        print(__doc__)
        os._exit(0)

    load_params()
    vsop = VSOP87d()
    sun = Sun()

    # Easter
    heappush(taskQueue, Task(HIGH_PRIORITY, doEaster, (start_year,)))

    # four equinox/solstice events
    for season in astronomia.globals.season_names:
        heappush(taskQueue, Task(HIGH_PRIORITY, doEquinox, (start_year, season)))

    # initialize rise-set-transit objects
    initRST(start_year)

    # start the task loop
    t = heappop(taskQueue)
    while t.jd < stop_jd:
        # apply(t.func, t.args)
        t.func(*t.args)
        t = heappop(taskQueue)


run()



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__init__.py for astronomia"""

__all__ = [
    "calendar",
    "constants",
    "coordinates",
    "dynamical",
    "moon",
    "equinox",
    "globals",
    "nutation",
    "riseset",
    "sun",
    "util",
    "planets",
]



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

import datetime
from unittest import TestCase

import numpy as np

from astronomia import lunar
from astronomia.calendar import cal_to_jd, hms_to_fday
from astronomia.util import r_to_d

_elp = lunar.Lunar()

# These are comparisons with Meuus, which aren't as good as my calculation
# compared against http://www.neoprogrammics.com/moon/
_comp_longitude = 133.162655
_comp_latitude = -3.229126
_comp_radius = 368409.7


class TestELP(TestCase):
    def test_dimension3(self):
        calc_longitude, calc_latitude, calc_radius = _elp.dimension3(2448724.5)
        calc_longitude = r_to_d(calc_longitude)
        calc_latitude = r_to_d(calc_latitude)

        np.testing.assert_array_almost_equal(calc_longitude, _comp_longitude, decimal=2)
        np.testing.assert_array_almost_equal(
            calc_longitude - _comp_longitude, -4.05 / 1000, decimal=0
        )
        np.testing.assert_array_almost_equal(calc_latitude, _comp_latitude, decimal=5)
        np.testing.assert_array_almost_equal(calc_radius, _comp_radius, decimal=1)

    def test_dimension3_01(self):
        calc_longitude, calc_latitude, calc_radius = _elp.dimension3(2456466.5)
        calc_longitude = r_to_d(calc_longitude)
        calc_latitude = r_to_d(calc_latitude)

        # from http://www.neoprogrammics.com/moon/
        np.testing.assert_array_almost_equal(calc_longitude, 264.8092609, decimal=5)
        np.testing.assert_array_almost_equal(calc_latitude, 3.2272014, decimal=5)
        np.testing.assert_array_almost_equal(calc_radius, 357206, decimal=0)

    def test_dimension(self):
        calc_longitude = _elp.dimension(2448724.5, "L")
        calc_longitude = r_to_d(calc_longitude)
        calc_latitude = _elp.dimension(2448724.5, "B")
        calc_latitude = r_to_d(calc_latitude)
        calc_radius = _elp.dimension(2448724.5, "R")

        np.testing.assert_array_almost_equal(calc_longitude, _comp_longitude, decimal=2)
        np.testing.assert_array_almost_equal(
            calc_longitude - _comp_longitude, -4.05 / 1000, decimal=0
        )
        np.testing.assert_array_almost_equal(calc_radius, _comp_radius, decimal=1)

    def test_arrays(self):
        calc_longitude = _elp.dimension([2448724.5, 2456466.5], "L")
        calc_longitude = r_to_d(calc_longitude)
        calc_latitude = _elp.dimension([2448724.5, 2456466.5], "B")
        calc_latitude = r_to_d(calc_latitude)
        calc_radius = _elp.dimension([2448724.5, 2456466.5], "R")

        np.testing.assert_array_almost_equal(
            calc_longitude, [_comp_longitude, 264.8092609], decimal=2
        )
        np.testing.assert_array_almost_equal(
            calc_longitude - _comp_longitude, [4.61 / 1000, 131600.0 / 1000], decimal=0
        )
        np.testing.assert_array_almost_equal(
            calc_latitude, [_comp_latitude, 3.2272014], decimal=5
        )
        np.testing.assert_array_almost_equal(
            calc_radius, [_comp_radius, 357206], decimal=1
        )

    def test_compare_to_schureman(self):
        rad2deg = 180.0 / np.pi
        dt = [datetime.datetime(i, 1, 1) for i in range(1800, 2001, 20)]
        dt = dt + [datetime.datetime(1990, 3, 3, 18, 0, 0)]
        dt = dt + [datetime.datetime(1999, 1, 10, 3, 0, 0)]
        jd = [
            cal_to_jd(i.year, i.month, i.day) + hms_to_fday(i.hour, i.minute, i.second)
            for i in dt
        ]
        jd = np.array(jd)
        Nv = np.mod(_elp.mean_longitude_ascending_node(jd) * rad2deg, 360)
        p = np.mod(_elp.mean_longitude_perigee(jd) * rad2deg, 360)
        s = np.mod(_elp.mean_longitude(jd) * rad2deg, 360)

        Nv_schureman = [
            33.25,
            6.47,
            339.64,
            312.81,
            285.98,
            259.16,
            232.38,
            205.55,
            178.72,
            151.89,
            125.07,
            318.45 - 3.12 - 0.11 - 0.04,
            144.39 + 0.00 - 0.48 - 0.01,
        ]
        p_schureman = [
            225.45,
            319.15,
            52.96,
            146.77,
            240.58,
            334.38,
            68.08,
            161.88,
            255.69,
            349.49,
            83.29,
            36.45 + 6.57 + 0.22 + 0.08,
            42.63 + 0.00 + 1.00 + 0.01,
        ]
        s_schureman = [
            342.31,
            102.71,
            236.29,
            9.87,
            143.45,
            277.03,
            37.43,
            171.01,
            304.59,
            78.16,
            211.74,
            np.mod(331.54 + 57.41 + 26.35 + 9.88, 360),
            82.36 + 0.00 + 118.59 + 1.65,
        ]
        for i in range(len(Nv)):
            np.testing.assert_array_almost_equal(Nv[i], Nv_schureman[i], decimal=1)
            np.testing.assert_array_almost_equal(p[i], p_schureman[i], decimal=1)
            np.testing.assert_array_almost_equal(s[i], s_schureman[i], decimal=1)



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Global values.

These can be set directly, or there is a routine astronomia.util.load_params()
which will assign them based on values in a parameter text file.

"""

#
# Abbreviation for standard timezone (e.g., "CST" for North American
# Central Standard Time)
#
standard_timezone_name = "UT"

#
# Time in fractional days to be subtracted from UT to calculate the standard
# time zone offset. Locations east of Greenwich should use negative values.
#
standard_timezone_offset = 0.0

#
# Abbreviation for daylight savings timezone (e.g., "CDT" for North American
# Central Daylight Time)
#
# This is optional. If set to None, no daylight savings conversions
# will be performed.
#
daylight_timezone_name = None

#
# Time in fractional days to be subtracted from UT to calculate the daylight
# savings time zone offset. Locations east of Greenwich should use negative
# values.
#
# This value is not used unless "daylight_timezone_name" has an value other
# than None.
#
daylight_timezone_offset = None

#
# Observer's longitude in radians, measured positive west of Greenwich,
# negative to the east. Should be between -pi...pi.
#
longitude = 0.0

#
# Observer's latitude in radians, measured positive north of the equator,
# negative to the south. Should be between -pi/2...pi/2.
#
latitude = 0.0

#
# Month names. There must be twelve. The default is three-character
# abbreviations so that listings line up.
#
month_names = (
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
)

#
# Season names. There must be four. These are used to characterize the
# equinoxes and solstices.
#
season_names = ("spring", "summer", "autumn", "winter")



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

import datetime
from unittest import TestCase

import numpy as np

from astronomia.calendar import cal_to_jd, hms_to_fday
from astronomia.constants import km_per_au
from astronomia.planets import VSOP87d, vsop_to_fk5
from astronomia.sun import (
    Sun,
    aberration_low,
    apparent_longitude_low,
    longitude_radius_low,
)
from astronomia.util import dms_to_d, r_to_d

sun = Sun()
vsop = VSOP87d()


class TestSun(TestCase):
    def test_longitude_radius_low(self):
        longitude, radius = longitude_radius_low(2448908.5)
        np.testing.assert_array_almost_equal(r_to_d(longitude), 199.90988, decimal=5)
        np.testing.assert_array_almost_equal(radius, 0.99766, decimal=5)

        longitude = apparent_longitude_low(2448908.5, longitude)
        np.testing.assert_array_almost_equal(r_to_d(longitude), 199.90895, decimal=5)

    def test_dimension3(self):
        L, B, R = sun.dimension3(2448908.5)
        np.testing.assert_array_almost_equal(
            r_to_d(L) * 3600, 199.907372 * 3600, decimal=0
        )
        np.testing.assert_array_almost_equal(r_to_d(B) * 3600, 0.644, decimal=1)
        np.testing.assert_array_almost_equal(
            R * km_per_au, 0.99760775 * km_per_au, decimal=-3
        )
        L, B = vsop_to_fk5(2448908.5, L, B)
        np.testing.assert_array_almost_equal(
            r_to_d(L) * 3600, 199.907347 * 3600, decimal=0
        )
        np.testing.assert_array_almost_equal(r_to_d(B) * 3600, 0.62, decimal=0)
        aberration = aberration_low(R)
        np.testing.assert_array_almost_equal(
            r_to_d(aberration) * 3600, -20.539, decimal=3
        )

        np.testing.assert_array_almost_equal(
            r_to_d(L) * 3600 * 100, dms_to_d(199, 54, 26.18) * 3600 * 100, decimal=1
        )
        np.testing.assert_array_almost_equal(
            r_to_d(B) * 3600 * 100, 0.72 * 100, decimal=1
        )
        np.testing.assert_array_almost_equal(R, 0.99760853)

    def test_compare_to_schureman(self):
        rad2deg = 180.0 / np.pi
        dt = [datetime.datetime(i, 1, 1) for i in range(1800, 2001, 20)]
        dt = dt + [datetime.datetime(1990, 3, 3, 18, 0, 0)]
        dt = dt + [datetime.datetime(1999, 1, 10, 3, 0, 0)]
        jd = [
            cal_to_jd(i.year, i.month, i.day) + hms_to_fday(i.hour, i.minute, i.second)
            for i in dt
        ]
        jd = np.array(jd)
        h = np.mod(sun.mean_longitude(jd) * rad2deg, 360)
        p1 = np.mod(sun.mean_longitude_perigee(jd) * rad2deg, 360)

        h_schureman = [
            280.41,
            279.57,
            279.73,
            279.88,
            280.04,
            280.19,
            279.36,
            279.51,
            279.67,
            279.82,
            279.97,
            280.39 + 58.15 + 1.97 + 0.74,
            280.21 + 0.00 + 8.87 + 0.12,
        ]
        p1_schureman = [
            279.50,
            279.85,
            280.19,
            280.53,
            280.88,
            281.22,
            281.56,
            281.91,
            282.25,
            282.60,
            282.94,
            282.77 + 0.00 + 0.00 + 0.00,
            282.92 + 0.00 + 0.00 + 0.00,
        ]
        for i in range(len(h)):
            np.testing.assert_array_almost_equal(h[i], h_schureman[i], decimal=1)
            np.testing.assert_array_almost_equal(p1[i], p1_schureman[i], decimal=1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

CGI interface to the Easter routine.

Generate dates of Easter for a range of years, both Gregorian
and Julian calendars.

One application handles both the input form and the displayed
results.

"""

import cgi
import sys
import time

#
# Rather than setup PYTHONPATH in a shell script, we add it
# through sys. That way we need only one file for the application.
#
# This path is specific to SourceForge.
#
sys.path.append("/home/groups/a/as/astronomia/lib/python")

from astronomia.calendar import easter

form = cgi.FieldStorage()

#
# Collect the input parameters
#


def display_form():
    fields = time.localtime(time.time())
    year = fields[0]

    print("Content-type: text/html")
    print()
    print("<HEAD>")
    print("<TITLE>Astronomia Easter Parameters</TITLE>")
    print("</HEAD>")

    print("<BODY>")
    print("<H1>Astronomia Easter Parameters</H1>")

    print(
        '<FORM action="http://astronomia.sourceforge.net/cgi-bin/easter-cgi.py" method="post">'
    )

    print("<TABLE>")
    print("<TR>")
    print('<TD><LABEL for="starting_year">Starting year</LABEL></TD>')
    print(
        '<TD><INPUT type="text" size=4 name="starting_year" id="starting_year" value="%d"></TD>'
        % (year - 5)
    )
    print("</TR>")
    print("<TR>")
    print('<TD><LABEL for="ending_year">Ending year</LABEL></TD>')
    print(
        '<TD><INPUT type="text" size=4 name="ending_year" id="ending_year" value="%d"></TD>'
        % (year + 5)
    )
    print("</TR>")
    print("</TABLE>")

    print(
        '<INPUT type="checkbox" checked name="gregorian_calendar" value="1"> Gregorian calendar<BR>'
    )
    print(
        '<INPUT type="checkbox" checked name="julian_calendar" value="1"> Julian calendar<BR>'
    )

    print('<INPUT type="submit" value="Run"> <INPUT type="reset">')

    print('<INPUT type="hidden" name="results" value="1">')

    print("</FORM>")
    print("</BODY>")


#
# Display an error message
#


def display_error(msg):
    print("<HEAD>")
    print("<TITLE>Astronomia Easter Error</TITLE>")
    print("</HEAD>")
    print("<BODY>")
    print("<H1>", msg, "</H1>")
    print("</BODY>")


#
# Display the results
#


def display_results():
    print("Content-type: text/html")
    print()
    print("<HEAD>")
    print("<TITLE>Astronomia Easter Results</TITLE>")
    print("</HEAD>")

    #
    # To do: We should validate the integer values
    #
    # If no starting_year, start with 5 years ago.
    #
    if "starting_year" in form:
        starting_year = int(form["starting_year"].value)
    else:
        fields = time.localtime(time.time())
        year = fields[0]
        starting_year = year - 5

    #
    # If no ending_year, end 10 years after the starting_year.
    #
    if "ending_year" in form:
        ending_year = int(form["ending_year"].value)
    else:
        ending_year = starting_year + 10

    #
    # Make sure end_year >= starting_year.
    #
    if ending_year < starting_year:
        display_error(
            "Ending year %d is earlier than starting year %d"
            % (ending_year, starting_year)
        )
        return

    #
    # We can't go farther back than the first Julian Day number.
    #
    if starting_year < -4712:
        display_error("Starting year %d is earlier than 4713BC" % starting_year)
        return

    julian_calendar = "julian_calendar" in form
    gregorian_calendar = "gregorian_calendar" in form

    #
    # Must have at least one calendar type.
    #
    if not julian_calendar and not gregorian_calendar:
        display_error("No calendar type selected")
        return

    #
    # Finally, we display the results.
    #
    print("<BODY>")
    print("<H1>Astronomia Easter Results</H1>")

    print("<TABLE BORDER>")

    print("<TR>")
    if gregorian_calendar:
        print("<TH>Gregorian</TH>")
    if julian_calendar:
        print("<TH>Julian</TH>")
    print("</TR>")

    for year in range(starting_year, ending_year + 1):
        print("<TR>")
        if gregorian_calendar:
            month, day = easter(year, True)
            print("<TD><TT>%02d-%02d-%d</TT></TD>" % (day, month, year))
        if julian_calendar:
            month, day = easter(year, False)
            print("<TD><TT>%02d-%02d-%d</TT></TD>" % (day, month, year))
        print("</TR>")
    print("</TABLE>")

    # handy diagnostic routines
    """
    print "<H1>print_environ</H1>"
    cgi.print_environ()
    print "<H1>print_directory</H1>"
    cgi.print_directory()
    print "<H1>print_environ_usage</H1>"
    cgi.print_environ_usage()
    print "<H1>FieldStorage</H1>"
    #form = cgi.FieldStorage()
    cgi.print_form(form)
    """

    print("</BODY>")


#
# Main routine. There is a hidden field "results" on the parameter form. If it
# is not present, we display the form. If it is, we display the results.
#
if "results" in form:
    display_results()
else:
    display_form()



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Geocentric solar position and radius, both low and high precision.

"""
import numpy as np

from . import globals as globls
from .calendar import jd_to_jcent
from .planets import VSOP87d
from .util import _scalar_if_one, d_to_r, dms_to_d, modpi2, polynomial


class Error(Exception):
    """Local exception class"""


class Sun:
    """High precision position calculations.

    This is a very light wrapper around the VSOP87d class. The geocentric
    longitude of the Sun is simply the heliocentric longitude of the Earth +
    180 degrees. The geocentric latitude of the Sun is the negative of the
    heliocentric latitude of the Earth. The radius is of course the same in
    both coordinate systems.

    """

    def __init__(self):
        self.vsop = VSOP87d()

    def mean_longitude(self, jd):
        """Return mean longitude.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - Longitude in radians

        """
        jd = np.atleast_1d(jd)
        T = jd_to_jcent(jd)

        # From astrolabe
        # X = polynomial((d_to_r(100.466457),
        #                d_to_r(36000.7698278),
        #                d_to_r(0.00030322),
        #                d_to_r(0.000000020)), T)

        # From AA, Naughter
        # Takes T/10.0
        X = polynomial(
            (
                d_to_r(100.4664567),
                d_to_r(360007.6982779),
                d_to_r(0.03032028),
                d_to_r(1.0 / 49931),
                d_to_r(-1.0 / 15300),
                d_to_r(-1.0 / 2000000),
            ),
            T / 10.0,
        )

        X = modpi2(X + np.pi)
        return _scalar_if_one(X)

    def mean_longitude_perigee(self, jd):
        """Return mean longitude of solar perigee.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - Longitude of solar perigee in radians

        """
        jd = np.atleast_1d(jd)
        T = jd_to_jcent(jd)

        X = polynomial((1012395.0, 6189.03, 1.63, 0.012), (T + 1)) / 3600.0
        X = d_to_r(X)

        X = modpi2(X)
        return _scalar_if_one(X)

    def dimension(self, jd, dim):
        """Return one of geocentric ecliptic longitude, latitude and radius.

        Arguments:
          - jd : Julian Day in dynamical time
          - dim : one of "L" (longitude) or "B" (latitude) or "R" (radius).

        Returns:
          - Either longitude in radians, or latitude in radians, or radius in
            au, depending on value of `dim`.

        """
        jd = np.atleast_1d(jd)
        X = self.vsop.dimension(jd, "Earth", dim)
        if dim == "L":
            X = modpi2(X + np.pi)
        elif dim == "B":
            X = -X
        return _scalar_if_one(X)

    def dimension3(self, jd):
        """Return geocentric ecliptic longitude, latitude and radius.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - longitude in radians
          - latitude in radians
          - radius in au

        """
        L = self.dimension(jd, "L")
        B = self.dimension(jd, "B")
        R = self.dimension(jd, "R")
        return L, B, R


#
# Constant terms
#
_kL0 = (d_to_r(280.46646), d_to_r(36000.76983), d_to_r(0.0003032))
_kM = (
    d_to_r(357.5291092),
    d_to_r(35999.0502909),
    d_to_r(-0.0001536),
    d_to_r(1.0 / 24490000),
)
_kC = (d_to_r(1.914602), d_to_r(-0.004817), d_to_r(-0.000014))

_ck3 = d_to_r(0.019993)
_ck4 = d_to_r(-0.000101)
_ck5 = d_to_r(0.000289)


def longitude_radius_low(jd):
    """Return geometric longitude and radius vector.

    Low precision. The longitude is accurate to 0.01 degree.  The latitude
    should be presumed to be 0.0. [Meeus-1998: equations 25.2 through 25.5

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - longitude in radians
      - radius in au

    """
    jd = np.atleast_1d(jd)
    T = jd_to_jcent(jd)
    L0 = polynomial(_kL0, T)
    M = polynomial(_kM, T)
    er = polynomial((0.016708634, -0.000042037, -0.0000001267), T)
    C = (
        polynomial(_kC, T) * np.sin(M)
        + (_ck3 - _ck4 * T) * np.sin(2 * M)
        + _ck5 * np.sin(3 * M)
    )
    L = modpi2(L0 + C)
    v = M + C
    R = 1.000001018 * (1 - er * er) / (1 + er * np.cos(v))
    return L, R


#
# Constant terms
#
_lk0 = d_to_r(125.04)
_lk1 = d_to_r(1934.136)
_lk2 = d_to_r(0.00569)
_lk3 = d_to_r(0.00478)


def apparent_longitude_low(jd, L):
    """Correct the geometric longitude for nutation and aberration.

    Low precision. [Meeus-1998: pg 164]

    Arguments:
      - `jd` : Julian Day in dynamical time
      - `L` : longitude in radians

    Returns:
      - corrected longitude in radians

    """
    jd = np.atleast_1d(jd)
    T = jd_to_jcent(jd)
    omega = _lk0 - _lk1 * T
    return _scalar_if_one(modpi2(L - _lk2 - _lk3 * np.sin(omega)))


#
# Constant terms
#
_lk4 = d_to_r(dms_to_d(0, 0, 20.4898))


def aberration_low(R):
    """Correct for aberration; low precision, but good enough for most uses.
    [Meeus-1998: pg 164]

    Arguments:
      - `R` : radius in au

    Returns:
      - correction in radians

    """
    return -_lk4 / R


def rise(
    year,
    month,
    day,
    longitude=globls.longitude,
    latitude=globls.latitude,
    gregorian=True,
):

    from .astronomia.constants import sun_rst_altitude

    jd = calendar.cal_to_jd(year, month, day, gregorian=gregorian)

    #
    # Sun
    #
    l, b, r = sun.dimension3(jd)

    # correct vsop coordinates
    l, b = vsop_to_fk5(jd, l, b)

    # nutation in longitude
    l += deltaPsi

    # aberration
    l += aberration_low(r)

    # equatorial coordinates
    ra, dec = ecl_to_equ(l, b, eps)

    obj = rstDict["Sun"]
    del obj.raList[0]
    del obj.decList[0]
    del obj.h0List[0]
    obj.raList.append(ra)
    obj.decList.append(dec)
    obj.h0List.append(sun_rst_altitude)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    """

"""Create a binary version of the VSOP87d database.

Usage:

    ./create_binary_vsop_db.py

"""
from astronomia.util import load_params
from astronomia.vsop87d import load_vsop87d_text_db, write_vsop87d_binary_db

load_params()
print("reading text database...")
load_vsop87d_text_db()
print("writing binary database...")
write_vsop87d_binary_db()



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Useful constants.

Don't change these unless you are moving to a new universe.

"""
from math import pi

#
# The number of radians in a circle.
#
pi2 = 2 * pi

#
# Ratio of Earth's polar to equitorial radius.
#
# flattening = 0.99664719
flattening = 0.99664713901477464

#
# Equitorial radius of the Earth in km.
# C_IERS
#
earth_equ_radius = 6378.1363

#
# How many minutes in a day?
#
minutes_per_day = 24.0 * 60.0

#
# How many days in minute?
#
days_per_minute = 1.0 / minutes_per_day

#
# How many seconds (time) in a day?
#
seconds_per_day = 24.0 * 60.0 * 60.0

#
# How many days in a second?
#
days_per_second = 1.0 / seconds_per_day

#
# How many kilometers in an astronomical unit?
#
# km_per_au = 149597870
# More accurate? Does it matter?
km_per_au = 149597870.691

#
# For rise-set-transit: altitude deflection caused by refraction
#
standard_rst_altitude = -0.00989078087105  # -0.5667 degrees
sun_rst_altitude = -0.0145438286569  # -0.8333 degrees



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

from unittest import TestCase

from astronomia.calendar import (
    cal_to_day_of_year,
    cal_to_jd,
    day_of_year_to_cal,
    jd_to_cal,
    jd_to_day_of_week,
)
from astronomia.coordinates import ecl_to_equ
from astronomia.util import d_to_r, r_to_d


class TestUtil(TestCase):
    def test_cal_to_jd(self):
        jd = cal_to_jd(1957, 10, 4.81)
        self.assertAlmostEqual(jd, 2436116.31, places=6)

        jd = cal_to_jd(333, 1, 27.5, False)
        self.assertAlmostEqual(jd, 1842713.0, places=6)

        jd0 = cal_to_jd(1910, 4, 20.0)
        jd1 = cal_to_jd(1986, 2, 9.0)
        self.assertEqual(jd1 - jd0, 27689)

        jd = cal_to_jd(1991, 7, 11)
        jd = jd + 10000
        yr, mo, day = jd_to_cal(jd)
        self.assertEqual(yr, 2018)
        self.assertEqual(mo, 11)
        self.assertEqual(day, 26)

    def test_dow(self):
        jd = cal_to_jd(1954, 6, 30.0)
        self.assertEqual(jd, 2434923.5)
        dow = jd_to_day_of_week(jd)
        self.assertEqual(dow, 3)

    def test_doy(self):
        N = cal_to_day_of_year(1978, 11, 14)
        self.assertEqual(N, 318)

        N = cal_to_day_of_year(1988, 4, 22)
        self.assertEqual(N, 113)

    def test_doy_to_cal(self):
        mo, day = day_of_year_to_cal(1978, 318)
        self.assertEqual(mo, 11)
        self.assertEqual(day, 14)

    def test_jd_to_cal(self):
        yr, mo, day = jd_to_cal(2436116.31)

        self.assertEqual(yr, 1957)
        self.assertEqual(mo, 10)
        self.assertAlmostEqual(day, 4.81)

        yr, mo, day = jd_to_cal(1842713.0, False)

        self.assertEqual(yr, 333)
        self.assertEqual(mo, 1)
        self.assertAlmostEqual(day, 27.5)

        yr, mo, day = jd_to_cal(1507900.13, False)

        self.assertEqual(yr, -584)
        self.assertEqual(mo, 5)
        self.assertAlmostEqual(day, 28.63)

    def test_ecl_to_equ(self):
        ra, dec = ecl_to_equ(d_to_r(113.215630), d_to_r(6.684170), d_to_r(23.4392911))
        self.assertAlmostEqual(r_to_d(ra), 116.328942, places=5)
        self.assertAlmostEqual(r_to_d(dec), 28.026183, places=6)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Usage:

    ./solstice.py start_year [stop_year]

Displays the instants of equinoxes and solstices for a range of years.
Times are accurate to one second.

The arguments must be integers.

If one argument is given, the display is for that year.

If two arguments are given, the display is for that range of
years.

"""

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""

import os
import sys

import astronomia.globals
from astronomia.calendar import lt_to_str, ut_to_lt
from astronomia.constants import days_per_second
from astronomia.dynamical import dt_to_ut
from astronomia.equinox import equinox, equinox_approx
from astronomia.util import load_params

tab = 4 * " "

if len(sys.argv) < 2:
    print(__doc__)
    os._exit(0)
elif len(sys.argv) < 3:
    start = int(sys.argv[1])
    stop = start
elif len(sys.argv) < 4:
    start = int(sys.argv[1])
    stop = int(sys.argv[2])
else:
    print(__doc__)
    os._exit(0)

load_params()
# sun = Sun() # load the database here so we don't get a message later

for yr in range(start, stop + 1):
    print(yr)
    for season in astronomia.globals.season_names:
        approx_jd = equinox_approx(yr, season)
        jd = equinox(approx_jd, season, days_per_second)
        ut = dt_to_ut(jd)
        lt, zone = ut_to_lt(ut)
        print(tab, season, lt_to_str(lt, zone))



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Collection of miscellaneous functions
    """

import numpy as np

from . import globals as globls
from .util import modpi2


class Error(Exception):
    """Local exception class"""

    pass


def ecl_to_equ(longitude, latitude, obliquity):
    """Convert ecliptic to equitorial coordinates.

    [Meeus-1998: equations 13.3, 13.4]

    Arguments:
      - `longitude` : ecliptic longitude in radians
      - `latitude` : ecliptic latitude in radians
      - `obliquity` : obliquity of the ecliptic in radians

    Returns:
      - Right accension in radians
      - Declination in radians

    """
    cose = np.cos(obliquity)
    sine = np.sin(obliquity)
    sinl = np.sin(longitude)
    ra = modpi2(np.arctan2(sinl * cose - np.tan(latitude) * sine, np.cos(longitude)))
    dec = np.arcsin(np.sin(latitude) * cose + np.cos(latitude) * sine * sinl)
    return ra, dec


def equ_to_horiz(H, decl):
    """Convert equitorial to horizontal coordinates.

    [Meeus-1998: equations 13.5, 13.6]

    Note that azimuth is measured westward starting from the south.

    This is not a good formula for using near the poles.

    Arguments:
      - `H` : hour angle in radians
      - `decl` : declination in radians

    Returns:
      - azimuth in radians
      - altitude in radians

    """
    cosH = np.cos(H)
    sinLat = np.sin(globls.latitude)
    cosLat = np.cos(globls.latitude)
    A = np.arctan2(np.sin(H), cosH * sinLat - np.tan(decl) * cosLat)
    h = np.arcsin(sinLat * np.sin(decl) + cosLat * np.cos(decl) * cosH)
    return A, h


def ell_to_geo(latitude, longitude, height):
    """Convert elliptic to geocentric coordinates.

    Arguments:
      - `latitude` : latitude
      - `longitude` : longitude
      - `height` : height

    Returns:
      - r
      - theta
      - phi

    """
    from .constants import earth_equ_radius

    f = 1.0 / 298.2564219846
    ea = earth_equ_radius / 1000

    ee = 2.0 * f - f * f

    sinLat = np.sin(latitude)
    cosLat = np.cos(latitude)

    N = ea / np.sqrt(1.0 - ee * sinLat * sinLat)

    Hx = (N + height) * cosLat
    Hy = (N * (1 - ee) + height) * sinLat

    r = np.sqrt(Hx * Hx + Hy * Hy)
    theta = np.arctan2(Hx, Hy)
    phi = longitude

    return (r, theta, phi)


def equ_to_ecl(ra, dec, obliquity):
    """Convert equitorial to ecliptic coordinates.

    [Meeus-1998: equations 13.1, 13.2]

    Arguments:
      - `ra` : right accension in radians
      - `dec` : declination in radians
      - `obliquity` : obliquity of the ecliptic in radians

    Returns:
      - ecliptic longitude in radians
      - ecliptic latitude in radians

    """
    cose = np.cos(obliquity)
    sine = np.sin(obliquity)
    sina = np.sin(ra)
    longitude = modpi2(np.arctan2(sina * cose + np.tan(dec) * sine, np.cos(ra)))
    latitude = modpi2(np.arcsin(np.sin(dec) * cose - np.cos(dec) * sine * sina))
    return longitude, latitude



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

from unittest import TestCase

from astronomia.util import interpolate3, interpolate_angle3


class TestUtil(TestCase):
    def test_interpolate3(self):
        y = interpolate3(0.18125, (0.884226, 0.877366, 0.870531))

        self.assertAlmostEqual(y, 0.876125, places=6)

    def test_interpolate_angle3(self):
        y = interpolate_angle3(0, (359, 0, 1))

        self.assertAlmostEqual(y, 0.0, places=6)



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

import math
from unittest import TestCase

import numpy as np

from astronomia.calendar import (
    cal_to_day_of_year,
    cal_to_jd,
    cal_to_jde,
    day_of_year_to_cal,
    easter,
    fday_to_hms,
    frac_yr_to_jd,
    is_leap_year,
    jd_to_cal,
    jd_to_day_of_week,
    sidereal_time_greenwich,
    yr_frac_mon_to_jd,
)
from astronomia.coordinates import ecl_to_equ
from astronomia.util import d_to_r, r_to_d


class TestUtil(TestCase):
    def test_cal_to_jd(self):
        jd = cal_to_jd(1957, 10, 4.81)
        self.assertAlmostEqual(jd, 2436116.31, places=6)

        jd = cal_to_jd(333, 1, 27.5, False)
        self.assertAlmostEqual(jd, 1842713.0, places=6)

        jd0 = cal_to_jd(1910, 4, 20.0)
        jd1 = cal_to_jd(1986, 2, 9.0)
        self.assertEqual(jd1 - jd0, 27689)

        jd = cal_to_jd(1991, 7, 11)
        jd = jd + 10000
        yr, mo, day = jd_to_cal(jd)
        self.assertEqual(yr, 2018)
        self.assertEqual(mo, 11)
        self.assertEqual(day, 26)

        jd = frac_yr_to_jd([1991 + 1.0 / 365.0, 1991])
        self.assertEqual(jd[0] - jd[1], 1.0)
        jd = frac_yr_to_jd([1991.5, 1991])
        self.assertEqual(jd[0] - jd[1], 182.5)

        jd = yr_frac_mon_to_jd(1991, [1.5, 1])
        self.assertEqual(jd[0] - jd[1], 15.5)

        self.assertRaises(ValueError, cal_to_jd, 1991, 13)
        self.assertRaises(ValueError, cal_to_jd, 1991, 12, 32)
        self.assertRaises(ValueError, cal_to_jd, 1991.1, 12.1)

    def test_fday_to_hms(self):
        self.assertEqual(fday_to_hms(0.5), (12, 0, 0))
        self.assertEqual(fday_to_hms(0.5006944444444444445), (12, 1, 0))
        self.assertEqual(
            fday_to_hms(0.5006944444444444445 + 0.0007060185185185186), (12, 2, 1)
        )

    def test_dow(self):
        jd = cal_to_jd(1954, 6, 30.0)
        self.assertEqual(jd, 2434923.5)
        dow = jd_to_day_of_week(jd)
        self.assertEqual(dow, 3)

    def test_sidereal(self):
        N = sidereal_time_greenwich(cal_to_jd(2004, 1, 1))
        testval = (6 * 3600 + 39 * 60 + 58.60298794778828) / 43200 * math.pi
        np.testing.assert_array_almost_equal(N, testval, decimal=4)
        N = sidereal_time_greenwich(cal_to_jd(2004, 1, [1, 2]))
        testval1 = (6 * 3600 + 43 * 60 + 55.15832794114431) / 43200 * math.pi
        np.testing.assert_array_almost_equal(N, [testval, testval1], decimal=4)

    def test_doy(self):
        N = cal_to_day_of_year(1978, 11, 14)
        self.assertEqual(N, 318)

        N = cal_to_day_of_year(1988, 4, 22)
        self.assertEqual(N, 113)

    def test_doy_to_cal(self):
        mo, day = day_of_year_to_cal(1978, 318)
        self.assertEqual(mo, 11)
        self.assertEqual(day, 14)

    def test_jd_to_cal(self):
        yr, mo, day = jd_to_cal(2436116.31)

        self.assertEqual(yr, 1957)
        self.assertEqual(mo, 10)
        self.assertAlmostEqual(day, 4.81)

        yr, mo, day = jd_to_cal(1842713.0, False)

        self.assertEqual(yr, 333)
        self.assertEqual(mo, 1)
        self.assertAlmostEqual(day, 27.5)

        yr, mo, day = jd_to_cal(1507900.13, False)

        self.assertEqual(yr, -584)
        self.assertEqual(mo, 5)
        self.assertAlmostEqual(day, 28.63)

    def test_jd_to_cal_array(self):
        yr, mo, day = jd_to_cal([1842713.0, 1507900.13], False)

        np.testing.assert_array_equal(yr, [333, -584])
        np.testing.assert_array_equal(mo, [1, 5])
        np.testing.assert_array_almost_equal(day, [27.5, 28.63])

    def test_ecl_to_equ(self):
        ra, dec = ecl_to_equ(d_to_r(113.215630), d_to_r(6.684170), d_to_r(23.4392911))
        self.assertAlmostEqual(r_to_d(ra), 116.328942, places=5)
        self.assertAlmostEqual(r_to_d(dec), 28.026183, places=6)

    def test_easter(self):
        tbl = (
            (1991, 3, 31),
            (1992, 4, 19),
            (1993, 4, 11),
            (1954, 4, 18),
            (2000, 4, 23),
            (1818, 3, 22),
        )

        for yr, mo, day in tbl:
            xmo, xday = easter(yr)
            self.assertEqual(xmo, mo)
            self.assertEqual(xday, day)

        for yr in [179, 711, 1243]:
            mo, day = easter(yr, False)
            self.assertEqual(mo, 4)
            self.assertEqual(day, 12)

    def test_cal_to_jde(self):
        tbl = [
            [(2013, 6, 18, 18, 25, 30), 2456462.267708],
            [(2013, 6, 18, 18, 26, 30), 2456462.268403],
        ]
        for date, testval in tbl:
            jd = cal_to_jde(*date)
            np.testing.assert_array_almost_equal(jd, testval, decimal=6)

    def test_is_leap_year(self):
        self.assertEqual(is_leap_year(2004), True)
        self.assertEqual(is_leap_year(2004, gregorian=False), True)



# -*- coding: utf-8 -*-
import os
import sys

import mando
import pandas as pd
from mando.rst_text_formatter import RSTHelpFormatter
from tstoolbox import tsutils
from tstoolbox.tstoolbox import createts


@mando.command(formatter_class=RSTHelpFormatter)
def right_ascension(
    latitude, longitude, body, input_ts=None, start_date=None, end_date=None, freq=None
):
    """Print out right ascension.

    :param body <str>:  Celestial body, one of ['sun', 'moon',
        'mercury', 'venus', 'mars', 'jupiter', 'saturn', 'uranus',
        'neptune', 'pluto']
    :param str input_ts:  Filename with data in 'ISOdate,value'
        format or '-' for stdin.
    :param str start_date:  The start_date of the series in
        ISOdatetime format, or 'None' for beginning.
    :param str end_date:  The end_date of the series in
        ISOdatetime format, or 'None' for end.
    :param freq:  To use this form --start_date and --end_date must be
        supplied also.  The pandas date offset code used to create the
        index.
    """
    usets = createts(
        input_ts=input_ts, start_date=start_date, end_date=end_date, freq=freq
    )
    pass


@mando.command(formatter_class=RSTHelpFormatter)
def risesettransit(
    latitude, longitude, start_date, end_date, body, h0=0, times="rise,set"
):
    """Print out rise, set, and/or transit times.

    :param latitude <float>:  Earth observer latitude in decimal degrees
    :param longitude <float>:  Earth observer longitude in decimal degrees
    :param start_date <str>:  ISO 8601 date string
    :param end_date <str>:  ISO 8601 date string
    :param body <str>:  Astronomical body, one of::

            Sun
            Moon
            Mercury
            Venus
            Mars
            Jupiter
            Saturn
            Uranus
            Neptune
            Pluto
    :param h0 <float>:  Earth observer standard altitude in radians
    :param times <str>: Comma separated list of times to include::

            "rise" for the rise time
            "set" for the set time
            "transit" for the transit time

        Defaults to "rise,set".
    """
    start_date = tsutils.parsedate(start_date)
    end_date = tsutils.parsedate(end_date)

    dr = pd.date_range(start=start_date, end=end_date).to_julian_date()
    start_jd = cal_to_jd(start_date)

    #
    # We need nutation values for each of three days
    #
    nutation = {}
    for day in (-1, 0, 1):
        jd = start_jd + day
        # nutation in longitude
        deltaPsi = nutation_in_longitude(jd)
        # apparent obliquity
        eps = obliquity(jd) + nutation_in_obliquity(jd)
        nutation[day] = deltaPsi, eps

    raList = []
    decList = []
    h0List = []
    for day in (-1, 0, 1):
        jd = start_jd + day
        deltaPsi, eps = nutation[day]
        if body == "Moon":
            l, b, r = moon.dimension3(jd)
            # nutation in longitude
            l += deltaPsi
            # equatorial coordinates
            ra, dec = ecl_to_equ(l, b, eps)
        elif body == "Sun":
            l, b, r = sun.dimension3(jd)
            # correct vsop coordinates
            l, b = vsop_to_fk5(jd, l, b)
            # nutation in longitude
            l += deltaPsi
            # aberration
            l += aberration_low(r)
            # equatorial coordinates
            ra, dec = ecl_to_equ(l, b, eps)
        else:
            ra, dec = geocentric_planet(jd, planet, deltaPsi, eps, days_per_second)


def main():
    if not os.path.exists("debug_astronomia"):
        sys.tracebacklimit = 0
    mando.main()


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Create a text version of the VSOP87d database.

Usage:

    ./create_text_vsop_db.py > vsop87d.txt   # or whatever path/file.

IMPORTANT: A text version of the database is provided with the Astronomia
package. You need to run this program ONLY if for some reason you want
to recreate that file.

This program does not use any routines from the Astronomia package. Run
it from a directory containing these files:

    VSOP87D.ear
    VSOP87D.jup
    VSOP87D.mar
    VSOP87D.mer
    VSOP87D.nep
    VSOP87D.sat
    VSOP87D.ura
    VSOP87D.ven

...which have been fetched from the ftp directory referenced at:

    http://cdsweb.u-strasbg.fr/cgi-bin/Cat?VI/81

The program will read each file, select the data required and combine all
of them into a format expected by the Astronomia vsop87d module.

Results are written to standard output; redirect them into a file in your
data directory and enter that path/file name into the astronomia_params.txt
file as the value of "vsop87d_text_path".

"""
#
# Here are file format notes from the original VSOP distribution.
#

# HEADER RECORD
# =============
# Specifications :
# - iv : code of VSOP87 version               integer     i1  col.18
# - bo : name of body                         character   a7  col.23-29
# - ic : index of coordinate                  integer     i1  col.42
# - it : degree alpha of time variable T      integer     i1  col.60
# - in : number of terms of series            integer     i7  col.61-67
#
# The code iv of the version is :
# iv = 4 for the version VSOP87D
#
# The names of the bodies are :
# MERCURY, VENUS, EARTH, MARS, JUPITER, SATURN, URANUS, NEPTUNE, SUN,
# and EMB for the Earth-Moon Barycenter.
#
# The index ic of the coordinates are :
# - for the spherical coordinates (versions B,D) :
#  1 : Longitude
#  2 : Latitude
#  3 : Radius
#
# The degree alpha of the time variable is equal to :
# 0 for periodic series, 1 to 5 for Poisson series.

# TERM RECORD
# ===========
# Specifications :
# iv : code of VSOP87 version                 integer     i1  col.02
# ic : index of coordinate                    integer     i1  col.04
# it : degree alpha of time variable T        integer     i1  col.05
# n  : rank of the term in a serie            integer     i5  col.06-10
# A  : amplitude A                            real dp f18.11  col.80-97
# B  : phase     B                            real dp f14.11  col.98-111
# C  : frequency C                            read dp f20.11  col.112-131

import string

planets = (
    "Mercury",
    "Venus",
    "Earth",
    "Mars",
    "Jupiter",
    "Saturn",
    "Uranus",
    "Neptune",
)
coords = ("L", "B", "R")

# each planet file...
for planet in planets:
    f = file("VSOP87D." + planet[:3].lower())
    s = f.readline()
    # header records...
    while s:
        assert s[17] == "4"  # model "d"
        assert (s[22:29]).rstrip() == planet.upper()  # planet name
        ic = int(s[41])  # coord type
        it = int(s[59])  # time degree
        nt = int(s[60:67])  # number of terms
        print(planet, coords[ic - 1], it, nt)
        # term records
        for i in range(nt):
            s = f.readline()
            assert s[1] == "4"  # model "d"
            ict = int(s[3])  # coord type
            assert ict == ic
            itt = int(s[4])  # time degree
            assert itt == it
            A = string.strip(s[79:97])
            B = string.strip(s[97:111])
            C = string.strip(s[111:131])
            print(A, B, C)
        s = f.readline()
    f.close()
#
# that's all
#



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

CGI interface to the Equinox/Solstice routines.

Will display a table of times of equinoxes and solstices for a
range of years.

One application handles both the input form and the displayed
results.

"""

import cgi
import sys
import time

#
# Rather than setup PYTHONPATH in a shell script, we add it
# through sys. That way we need only one file for the application.
#
#
# This path is specific to SourceForge.
#
sys.path.append("/home/groups/a/as/astronomia/lib/python")

import astronomia.globals
from astronomia.calendar import lt_to_str, ut_to_lt
from astronomia.constants import days_per_second
from astronomia.dynamical import dt_to_ut
from astronomia.equinox import equinox, equinox_approx

standard_zones = (
    "GMT-12",
    "GMT-11",
    "GMT-10",
    "GMT-9",
    "PST",
    "MST",
    "CST",
    "EST",
    "GMT-4",
    "GMT-3",
    "GMT-2",
    "GMT-1",
    "GMT",
    "GMT+1",
    "GMT+2",
    "GMT+3",
    "GMT+4",
    "GMT+5",
    "GMT+6",
    "GMT+7",
    "GMT+8",
    "GMT+9",
    "GMT+10",
    "GMT+11",
    "GMT+12",
)

standard_zone_hours = {
    "GMT-12": -12,
    "GMT-11": -11,
    "GMT-10": -10,
    "GMT-9": -9,
    "PST": -8,
    "MST": -7,
    "CST": -6,
    "EST": -5,
    "GMT-4": -4,
    "GMT-3": -3,
    "GMT-2": -2,
    "GMT-1": -1,
    "GMT": 0,
    "GMT+1": 1,
    "GMT+2": 2,
    "GMT+3": 3,
    "GMT+4": 4,
    "GMT+5": 5,
    "GMT+6": 6,
    "GMT+7": 7,
    "GMT+8": 8,
    "GMT+9": 9,
    "GMT+10": 10,
    "GMT+11": 11,
    "GMT+12": 12,
}

daylight_zones = (
    "GMT-12",
    "GMT-11",
    "GMT-10",
    "GMT-9",
    "GMT-8",
    "PDT",
    "MDT",
    "CDT",
    "EDT",
    "GMT-3",
    "GMT-2",
    "GMT-1",
    "GMT",
    "GMT+1",
    "GMT+2",
    "GMT+3",
    "GMT+4",
    "GMT+5",
    "GMT+6",
    "GMT+7",
    "GMT+8",
    "GMT+9",
    "GMT+10",
    "GMT+11",
    "GMT+12",
)

daylight_zone_hours = {
    "GMT-12": -12,
    "GMT-11": -11,
    "GMT-10": -10,
    "GMT-9": -9,
    "GMT-8": -8,
    "PDT": -7,
    "MDT": -6,
    "CDT": -5,
    "EDT": -4,
    "GMT-3": -3,
    "GMT-2": -2,
    "GMT-1": -1,
    "GMT": 0,
    "GMT+1": 1,
    "GMT+2": 2,
    "GMT+3": 3,
    "GMT+4": 4,
    "GMT+5": 5,
    "GMT+6": 6,
    "GMT+7": 7,
    "GMT+8": 8,
    "GMT+9": 9,
    "GMT+10": 10,
    "GMT+11": 11,
    "GMT+12": 12,
}

default_standard_zone = "CST"
default_daylight_zone = "CDT"

form = cgi.FieldStorage()

#
# Collect the input parameters
#


def display_form():
    fields = time.localtime(time.time())
    year = fields[0]

    print("Content-type: text/html")
    print()
    print("<HEAD>")
    print("<TITLE>Astronomia Solstice Parameters</TITLE>")
    print("</HEAD>")

    print("<BODY>")
    print("<H1>Astronomia Solstice Parameters</H1>")

    print(
        '<FORM action="http://astronomia.sourceforge.net/cgi-bin/solstice-cgi.py" method="post">'
    )

    print("<TABLE>")
    print("<TR>")
    print('<TD><LABEL for="starting_year">Starting year</LABEL></TD>')
    print(
        '<TD><INPUT type="text" size=4 name="starting_year" id="starting_year" value="%d"></TD>'
        % (year - 5)
    )
    print("</TR>")
    print("<TR>")
    print('<TD><LABEL for="ending_year">Ending year</LABEL></TD>')
    print(
        '<TD><INPUT type="text" size=4 name="ending_year" id="ending_year" value="%d"></TD>'
        % (year + 5)
    )
    print("</TR>")
    print("<TR>")
    print('<TD><LABEL for="standard_zone">Standard Time Zone</LABEL></TD>')
    print('<TD><SELECT name = "standard_zone">')
    for zone in standard_zones:
        if zone == default_standard_zone:
            print("<OPTION SELECTED>" + zone + "</OPTION>")
        else:
            print("<OPTION>" + zone + "</OPTION>")
    print("</SELECT></TD>")
    print("</TR>")
    print("<TR>")
    print('<TD><LABEL for="daylight_zone">Daylight Time Zone</LABEL></TD>')
    print('<TD><SELECT name = "daylight_zone">')
    for zone in daylight_zones:
        if zone == default_daylight_zone:
            print("<OPTION SELECTED>" + zone + "</OPTION>")
        else:
            print("<OPTION>" + zone + "</OPTION>")
    print("</SELECT></TD>")
    print("</TR>")
    print("</TABLE>")

    print('<INPUT type="submit" value="Run"> <INPUT type="reset">')

    print('<INPUT type="hidden" name="results" value="1">')

    print("</FORM>")
    print("</BODY>")


#
# Display an error message
#


def display_error(msg):
    print("<HEAD>")
    print("<TITLE>Astronomia Solstice Error</TITLE>")
    print("</HEAD>")
    print("<BODY>")
    print("<H1>", msg, "</H1>")
    print("</BODY>")


#
# Display the results
#


def display_results():
    #
    # Assign the global parameters directly so we don't have to read a param file.
    #
    astronomia.globals.standard_timezone_name = form["standard_zone"].value
    astronomia.globals.standard_timezone_offset = (
        -standard_zone_hours[astronomia.globals.standard_timezone_name] / 24.0
    )
    astronomia.globals.daylight_timezone_name = form["daylight_zone"].value
    astronomia.globals.daylight_timezone_offset = (
        -daylight_zone_hours[astronomia.globals.daylight_timezone_name] / 24.0
    )
    astronomia.globals.vsop87d_text_path = (
        "/home/groups/a/as/astronomia/data/vsop87d.txt"
    )
    astronomia.globals.vsop87d_binary_path = (
        "/home/groups/a/as/astronomia/data/vsop87d.dat"
    )

    print("Content-type: text/html")
    print()
    print("<HEAD>")
    print("<TITLE>Astronomia Solstice Results</TITLE>")
    print("</HEAD>")

    #
    # To do: We should validate the integer values
    #
    # If no starting_year, start with 5 years ago.
    #
    if "starting_year" in form:
        starting_year = int(form["starting_year"].value)
    else:
        fields = time.localtime(time.time())
        year = fields[0]
        starting_year = year - 5

    #
    # If no ending_year, end 10 years after the starting_year.
    #
    if "ending_year" in form:
        ending_year = int(form["ending_year"].value)
    else:
        ending_year = starting_year + 10

    #
    # Make sure end_year >= starting_year.
    #
    if ending_year < starting_year:
        display_error(
            "Ending year %d is earlier than starting year %d"
            % (ending_year, starting_year)
        )
        return

    #
    # We can't go farther back than the first Julian Day number.
    #
    if starting_year < -4712:
        display_error("Starting year %d is earlier than 4713BC" % starting_year)
        return

    #
    # Finally, we display the results.
    #
    print("<BODY>")
    print("<H1>Astronomia Solstice Results</H1>")

    print("<TABLE BORDER>")

    print("<TR>")
    print("<TH>Vernal Equinox</TH>")
    print("<TH>Summer Solstice</TH>")
    print("<TH>Autumnal Equinox</TH>")
    print("<TH>Winter Solstice</TH>")
    print("</TR>")

    for year in range(starting_year, ending_year + 1):
        print("<TR>")
        for season in astronomia.globals.season_names:
            approx_jd = equinox_approx(year, season)
            jd = equinox(approx_jd, season, days_per_second)
            ut = dt_to_ut(jd)
            lt, zone = ut_to_lt(ut)
            print("<TD>" + lt_to_str(lt, zone) + "</TD>")
        print("</TR>")
    print("</TABLE>")

    # handy diagnostic routines
    """
    print "<H1>print_environ</H1>"
    cgi.print_environ()
    print "<H1>print_directory</H1>"
    cgi.print_directory()
    print "<H1>print_environ_usage</H1>"
    cgi.print_environ_usage()
    print "<H1>FieldStorage</H1>"
    #form = cgi.FieldStorage()
    cgi.print_form(form)
    """

    print("</BODY>")


#
# Main routine. There is a hidden field "results" on the parameter form. If it
# is not present, we display the form. If it is, we display the results.
#
if "results" in form:
    display_results()
else:
    display_form()



# -*- coding: utf-8 -*-
"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Useful constants terms.

Don't change these unless you are moving to a new universe.

"""
from .util import d_to_r

#
# Constant terms.
#
kL1 = (
    d_to_r(218.3164477),
    d_to_r(481267.88123421),
    d_to_r(-0.0015786),
    d_to_r(1.0 / 538841),
    d_to_r(-1.0 / 65194000),
)
kD = (
    d_to_r(297.8501921),
    d_to_r(445267.1114034),
    d_to_r(-0.0018819),
    d_to_r(1.0 / 545868),
    d_to_r(-1.0 / 113065000),
)
kM = (
    d_to_r(357.5291092),
    d_to_r(35999.0502909),
    d_to_r(-0.0001536),
    d_to_r(1.0 / 24490000),
)
kM1 = (
    d_to_r(134.9633964),
    d_to_r(477198.8675055),
    d_to_r(0.0087414),
    d_to_r(1.0 / 69699),
    d_to_r(-1.0 / 14712000),
)
kF = (
    d_to_r(93.2720950),
    d_to_r(483202.0175233),
    d_to_r(-0.0036539),
    d_to_r(-1.0 / 3526000),
    d_to_r(1.0 / 863310000),
)
ko = (
    d_to_r(125.0445479),
    d_to_r(-1934.1362891),
    d_to_r(0.0020754),
    d_to_r(1.0 / 467441.0),
    d_to_r(1.0 / 60616000.0),
)



# -*- coding: utf-8 -*-
"""
Tests for the elp2000 functions.
"""

import os.path
from unittest import TestCase

import numpy as np

from astronomia.calendar import hms_to_fday
from astronomia.constants import days_per_second, km_per_au, pi2
from astronomia.planets import VSOP87d, geocentric_planet
from astronomia.util import d_to_r, dms_to_d, r_to_d

vsop = VSOP87d()


class TestVSOP(TestCase):
    def test_dimension3(self):
        L, B, R = vsop.dimension3(2448976.5, "Venus")
        np.testing.assert_array_almost_equal(
            r_to_d(L) * 3600, 26.11428 * 3600, decimal=0
        )
        np.testing.assert_array_almost_equal(
            r_to_d(B) * 3600, -2.62070 * 3600, decimal=0
        )
        np.testing.assert_array_almost_equal(
            R * km_per_au, 0.724603 * km_per_au, decimal=-4
        )

    def test_geocentric_planet(self):
        ra, dec = geocentric_planet(
            2448976.5,
            "Venus",
            d_to_r(dms_to_d(0, 0, 16.749)),
            d_to_r(23.439669),
            days_per_second,
        )
        np.testing.assert_array_almost_equal(
            r_to_d(ra), r_to_d(hms_to_fday(21, 4, 41.454) * pi2), decimal=5
        )
        np.testing.assert_almost_equal(r_to_d(dec), dms_to_d(-18, 53, 16.84), decimal=5)


class TestVSOPDatabase(TestCase):
    def test_vsop87d_chk(self):
        """
        where "vsop87.chk" has been fetched from the ftp directory referenced
        at:

            http://cdsweb.u-strasbg.fr/cgi-bin/Cat?VI/81

        The program scans through the file and selects those 80 tests which
        apply to VSOP87d. We calculate results for each test and compare
        with the given value.

        Result: all calculations match within 1e-10 radians or au.

        """
        refs = []
        bname = os.path.dirname(__file__)
        f = open(os.path.join(bname, "vsop87.chk"), "r")
        line = f.readline()
        while line:
            fields = line.split()
            if fields:
                if fields[0] == "VSOP87D":
                    planet = fields[1]
                    planet = planet[0] + planet[1:].lower()
                    jd = fields[2]
                    jd = float(jd[2:])
                    line = f.readline()
                    fields = line.split()
                    l = float(fields[1])
                    b = float(fields[4])
                    r = float(fields[7])
                    refs.append((planet, jd, l, b, r))
            line = f.readline()
        f.close()

        for planet, jd, l, b, r in refs:
            L, B, R = vsop.dimension3(jd, planet)
            np.testing.assert_array_almost_equal(L, l)
            np.testing.assert_array_almost_equal(B, b)
            np.testing.assert_array_almost_equal(R, r)



# -*- coding: utf-8 -*-

import os
import sys

from setuptools import setup

pkg_name = "astronomia"

version = open("VERSION").readline().strip()

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist")
    os.system("twine upload dist/{pkg_name}-{version}.tar.gz".format(**locals()))
    sys.exit()

README = open("README.rst").read()

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
    "tstoolbox >= 103",
]

setup(
    name=pkg_name,
    version=version,
    description="Library for calculation of ephemeris and other astronomical calculations",
    long_description=README,
    classifiers=[
        # Get strings from
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords="ephemeris astronomy",
    author="Tim Cera, P.E.",
    author_email="tim@cerazone.net",
    url="http://timcera.bitbucket.io/astronomia/docsrc/index.html",
    license="GPL2",
    packages=["astronomia"],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={"console_scripts": ["astronomia=astronomia.astronomia:main"]},
    scripts=[
        "apps/solstice.py",
        "apps/check_perihelion.py",
        "apps/cronus.py",
        "apps/easter-cgi.py",
        "apps/solstice-cgi.py",
    ],
    data_files=[
        (
            os.path.join("share", "astronomia"),
            [os.path.join("apps", "astronomia_params.txt")],
        )
    ],
    test_suite="tests",
    python_requires=">=3.6",
)
